# pyOCD debugger
# Copyright (c) 2021 NXP
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ..family.target_lpc5500 import LPC5500Family
from ...core.memory_map import (FlashRegion, RamRegion, RomRegion, MemoryMap)
from ...debug.svd.loader import SVDFile

# Note: the DFP has both S and NS flash algos, but they are exactly the same except for the address range.
FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe00abe00,
    0xf240b580, 0xf2c00004, 0xf6420000, 0xf84961e0, 0xf2401000, 0xf2c52000, 0x21000000, 0x1080f8c0,
    0x1084f8c0, 0x1180f8c0, 0x71fbf647, 0xf6406001, 0x21ff6004, 0x0000f2c5, 0x01def2cc, 0xf04f6001,
    0x210240a0, 0xf2407001, 0xf2c0000c, 0x44480000, 0xf874f000, 0xbf182800, 0xbd802001, 0x47702000,
    0xf240b580, 0xf2c0000c, 0xf2460000, 0x4448636c, 0xf6c62100, 0xf44f3365, 0xf0003260, 0x2800f86d,
    0x2001bf18, 0xbf00bd80, 0xf020b580, 0xf2404170, 0xf2c0000c, 0xf2460000, 0x4448636c, 0x3365f6c6,
    0x4200f44f, 0xf858f000, 0xbf182800, 0xbd802001, 0xb081b5f0, 0x070cf240, 0x460d4614, 0xf0200441,
    0xf2c04670, 0xd10a0700, 0x636cf246, 0x0007eb09, 0xf6c64631, 0xf44f3365, 0xf0004200, 0xf5b5f83d,
    0xbf987f00, 0x7500f44f, 0x0007eb09, 0x46224631, 0xf000462b, 0x2800f847, 0x2001bf18, 0xbdf0b001,
    0x460cb5b0, 0xf0204605, 0x46114070, 0xf0004622, 0x2800f8b8, 0x4425bf08, 0xbdb04628, 0x460ab580,
    0x4170f020, 0x000cf240, 0x0000f2c0, 0xf0004448, 0x2800f83f, 0x2001bf18, 0x0000bd80, 0x0108f240,
    0x0100f2c0, 0xf8092201, 0xf64f2001, 0xf2c101dc, 0x68093102, 0xbf004708, 0x0c08f240, 0x0c00f2c0,
    0xc00cf819, 0x0f00f1bc, 0xf64abf07, 0xf2c13c4f, 0xf64f3c00, 0xf2c10ce0, 0xbf183c02, 0xc000f8dc,
    0xbf004760, 0x0c08f240, 0x0c00f2c0, 0xc00cf819, 0x0f00f1bc, 0xf248bf07, 0xf2c17c9b, 0xf64f3c02,
    0xf2c10ce4, 0xbf183c02, 0xc000f8dc, 0xbf004760, 0x0308f240, 0x0300f2c0, 0x3003f819, 0xbf072b00,
    0x3381f64a, 0x3300f2c1, 0x03e8f64f, 0x3302f2c1, 0x681bbf18, 0xbf004718, 0x0c08f240, 0x0c00f2c0,
    0xc00cf819, 0x0f00f1bc, 0xf64abf07, 0xf2c14ca5, 0xf64f3c00, 0xf2c10cec, 0xbf183c02, 0xc000f8dc,
    0xbf004760, 0x03f0f64f, 0x3302f2c1, 0x4718681b, 0x01f4f64f, 0x3102f2c1, 0x47086809, 0x01f8f64f,
    0x3102f2c1, 0x47086809, 0x03fcf64f, 0x3302f2c1, 0x4718681b, 0x1c04f64f, 0x3c02f2c1, 0xc000f8dc,
    0xbf004760, 0x1208f64f, 0x3202f2c1, 0x47106812, 0x120cf64f, 0x3202f2c1, 0x47106812, 0x1310f64f,
    0x3302f2c1, 0x4718681b, 0x1200f64f, 0x3202f2c1, 0x47106812, 0x1c18f64f, 0x3c02f2c1, 0xc000f8dc,
    0xea404760, 0xb5100301, 0xd10f079b, 0xd30d2a04, 0xc908c810, 0x429c1f12, 0xba20d0f8, 0x4288ba19,
    0x2001d901, 0xf04fbd10, 0xbd1030ff, 0x07d3b11a, 0x1c52d003, 0x2000e007, 0xf810bd10, 0xf8113b01,
    0x1b1b4b01, 0xf810d107, 0xf8113b01, 0x1b1b4b01, 0x1e92d101, 0x4618d1f1, 0x0000bd10, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000,
    0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000005,
    'pc_unInit': 0x20000061,
    'pc_program_page': 0x200000b5,
    'pc_erase_sector': 0x2000008d,
    'pc_eraseAll': 0x20000065,

    'static_base' : 0x20000000 + 0x00000004 + 0x000002dc,
    'begin_stack' : 0x20000500,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x200,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20001000, 0x20001200],   # Enable double buffering
    'min_program_length' : 0x200,

    # Relative region addresses and sizes
    'ro_start': 0x0,
    'ro_size': 0x2dc,
    'rw_start': 0x2dc,
    'rw_size': 0x4,
    'zi_start': 0x2e0,
    'zi_size': 0x44,

    # Flash information
    'flash_start': 0x0,
    'flash_size': 0x40000,
    'sector_sizes': (
        (0x0, 0x8000),
    )
}


class LPC55S36(LPC5500Family):

    MEMORY_MAP = MemoryMap(
        FlashRegion(name='nsflash',     start=0x00000000, length=0x040000, access='rx',
            page_size=0x200,
            sector_size=0x8000,
            is_boot_memory=True,
            are_erased_sectors_readable=False,
            algo=FLASH_ALGO),
        RomRegion(  name='nsrom',       start=0x03000000, length=0x020000, access='rx'),
        RamRegion(  name='nscoderam',   start=0x04000000, length=0x4000, access='rwx',
            default=False),
        FlashRegion(name='sflash',      start=0x10000000, length=0x040000, access='rx',
            page_size=0x200,
            sector_size=0x8000,
            is_boot_memory=True,
            are_erased_sectors_readable=False,
            algo=FLASH_ALGO,
            alias='nsflash'),
        RomRegion(  name='srom',        start=0x13000000, length=0x020000, access='srx',
            alias='nsrom'),
        RamRegion(  name='scoderam',    start=0x14000000, length=0x4000, access='srwx',
            alias='nscoderam',
            default=False),
        RamRegion(  name='nsram',       start=0x20000000, length=0x01c000, access='rwx'),
        RamRegion(  name='sram',        start=0x30000000, length=0x01c000, access='srwx',
            alias='nsram'),
        )

    def __init__(self, session):
        super(LPC55S36, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("LPC55S36.xml")



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (RamRegion, MemoryMap)

class AN522(CoreSightTarget):

    VENDOR = "Arm"
    
    MEMORY_MAP = MemoryMap(
        RamRegion(  name='itcm',        start=0x00000000, length=0x00080000, access='rwx'),
        RamRegion(  name='sram',        start=0x30000000, length=0x00100000, access='rwx'),
        RamRegion(  name='dram6_ns',    start=0x60000000, length=0x10000000, access='rwx'),
        RamRegion(  name='dram7_s',     start=0x70000000, length=0x10000000, access='rwxs'),
        RamRegion(  name='dram8_ns',    start=0x80000000, length=0x10000000, access='rwx'),
        RamRegion(  name='dram9_s',     start=0x90000000, length=0x10000000, access='rwxs'),
        )

    def __init__(self, session):
        super(AN522, self).__init__(session, self.MEMORY_MAP)




# pyOCD debugger
# Copyright (c) 2018 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile

class DBGMCU:
    CR = 0xE0042004
    CR_VALUE = 0x7 # DBG_STANDBY | DBG_STOP | DBG_SLEEP

    APB1_FZ = 0xE0042008
    APB1_FZ_VALUE = 0x07E01DFF

    APB2_FZ = 0xE004200C
    APB2_FZ_VALUE = 0x00070003


FLASH_ALGO = { 'load_address' : 0x20000000,
               'instructions' : [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x03004601, 0x28200e00, 0x0940d302, 0xe0051d00, 0xd3022810, 0x1cc00900, 0x0880e000, 0xd50102c9,
    0x43082110, 0x48424770, 0x60414940, 0x60414941, 0x60012100, 0x22f068c1, 0x60c14311, 0x06806940,
    0x483ed406, 0x6001493c, 0x60412106, 0x6081493c, 0x47702000, 0x69014836, 0x43110542, 0x20006101,
    0xb5104770, 0x69014832, 0x43212404, 0x69016101, 0x431103a2, 0x49336101, 0xe0004a30, 0x68c36011,
    0xd4fb03db, 0x43a16901, 0x20006101, 0xb530bd10, 0xffb6f7ff, 0x68ca4926, 0x431a23f0, 0x240260ca,
    0x690a610c, 0x0e0006c0, 0x610a4302, 0x03e26908, 0x61084310, 0x4a214823, 0x6010e000, 0x03ed68cd,
    0x6908d4fb, 0x610843a0, 0x060068c8, 0xd0030f00, 0x431868c8, 0x200160c8, 0xb570bd30, 0x1cc94d14,
    0x68eb0889, 0x26f00089, 0x60eb4333, 0x612b2300, 0xe0174b15, 0x431c692c, 0x6814612c, 0x68ec6004,
    0xd4fc03e4, 0x0864692c, 0x612c0064, 0x062468ec, 0xd0040f24, 0x433068e8, 0x200160e8, 0x1d00bd70,
    0x1f091d12, 0xd1e52900, 0xbd702000, 0x45670123, 0x40023c00, 0xcdef89ab, 0x00005555, 0x40003000,
    0x00000fff, 0x0000aaaa, 0x00000201, 0x00000000
    ],

    'pc_init' : 0x20000047,
    'pc_unInit': 0x20000075,
    'pc_program_page': 0x200000fb,
    'pc_erase_sector': 0x200000af,
    'pc_eraseAll' : 0x20000083,

    'static_base' : 0x20000000 + 0x00000020 + 0x0000014c,
    'begin_stack' : 0x20000000 + 0x00000800,
    'begin_data' : 0x20003000,
    'page_buffers' : [0x20003000, 0x20004000],
    'min_program_length' : 2,
    'analyzer_supported' : True,
    'analyzer_address' : 0x20002000
  }

class STM32F412xE(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x08000000, length=0x10000, sector_size=0x4000,
                        page_size=0x1000,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        FlashRegion( start=0x08010000, length=0x10000, sector_size=0x10000,
                        page_size=0x1000,
                        algo=FLASH_ALGO),
        FlashRegion( start=0x08020000, length=0x60000, sector_size=0x20000,
                        page_size=0x1000,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x40000)
        )

    def __init__(self, session):
        super(STM32F412xE, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("STM32F41x.svd")

    def post_connect_hook(self):
        self.write32(DBGMCU.CR, DBGMCU.CR_VALUE)
        self.write32(DBGMCU.APB1_FZ, DBGMCU.APB1_FZ_VALUE)
        self.write32(DBGMCU.APB2_FZ, DBGMCU.APB2_FZ_VALUE)

class STM32F412xG(CoreSightTarget):

    VENDOR = "STMicroelectronics"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x08000000, length=0x10000, sector_size=0x4000,
                        page_size=0x1000,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        FlashRegion( start=0x08010000, length=0x10000, sector_size=0x10000,
                        page_size=0x1000,
                        algo=FLASH_ALGO),
        FlashRegion( start=0x08020000, length=0xE0000, sector_size=0x20000,
                        page_size=0x1000,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x40000)
        )

    def __init__(self, session):
        super(STM32F412xG, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("STM32F41x.svd")

    def post_connect_hook(self):
        self.write32(DBGMCU.CR, DBGMCU.CR_VALUE)
        self.write32(DBGMCU.APB1_FZ, DBGMCU.APB1_FZ_VALUE)
        self.write32(DBGMCU.APB2_FZ, DBGMCU.APB2_FZ_VALUE)



# pyOCD debugger
# Copyright (c) 2018-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
import logging
from pyocd.core.target import Target
from time import time
from typing import (Any, Callable, Dict, List, Optional, TYPE_CHECKING, Union)

from .builder import (
    FlashBuilder,
    MemoryBuilder,
    ProgrammingInfo,
    get_page_count,
    get_sector_count,
)
from ..core import exceptions
from ..utility.progress import print_progress

if TYPE_CHECKING:
    from ..core.session import Session
    from ..core.memory_map import (
        MemoryRegion,
        MemoryMap,
        RamRegion,
    )

LOG = logging.getLogger(__name__)

ProgressCallback = Callable[[Union[int, float]], None]

@dataclass
class DataChunk:
    addr: int
    data: Union[bytes, bytearray]

class RamBuilder(MemoryBuilder):
    """@brief Memory builder for writing potentially discontiguous data to RAM."""

    ## Maximum number of bytes to write at once. This is primarily done so progress is updated occasionally.
    _MAX_WRITE_SIZE = 4096

    def __init__(self, session: "Session", region: "RamRegion") -> None:
        """@brief Constructor."""
        super().__init__()
        self._session = session
        self._region = region
        self._chunks: List[DataChunk] = []

    def add_data(self, addr: int, data: Union[bytes, bytearray]) -> None:
        # Make sure this address range is contained by our region.
        if not self._region.contains_range(start=addr, length=len(data)):
            raise ValueError(f"Attempt to add data ({addr:#010x}-{addr + len(data) - 1:#010x}) outside "
                              "of RAM builder region {self._region}")

        self._chunks.append(DataChunk(addr, bytearray(data)))
        self._chunks.sort(key=lambda c: c.addr)
        self._buffered_data_size += len(data)

    def program(self, progress_cb: Optional[ProgressCallback] = None, **kwargs: Any) -> ProgrammingInfo:
        target = self._session.target
        assert isinstance(target, Target)

        if progress_cb is not None:
            progress_cb(0.0)

        written_byte_count = 0
        start_time = time()

        for chunk in self._chunks:
            chunk_size = len(chunk.data)
            offset_within_chunk = 0
            while offset_within_chunk < chunk_size:
                write_size = min(self._MAX_WRITE_SIZE, chunk_size - offset_within_chunk)
                target.write_memory_block8(
                            chunk.addr + offset_within_chunk,
                            chunk.data[offset_within_chunk:offset_within_chunk + write_size]
                            )

                offset_within_chunk += write_size
                written_byte_count += write_size
                if progress_cb is not None:
                    progress_cb(written_byte_count / self._buffered_data_size)

        # Make sure progress has reached 100%.
        if progress_cb is not None:
            progress_cb(1.0)

        # Return some performance numbers.
        return ProgrammingInfo(
            program_time=time() - start_time,
            total_byte_count=self.buffered_data_size,
            program_byte_count=self.buffered_data_size,
            )

    @property
    def region(self) -> "MemoryRegion":
        return self._region


class MemoryLoader:
    """@brief Handles high level programming of raw binary data to memory.
    
    If you need file programming, either binary files or other formats, please see the
    FileProgrammer class.
    
    This manager provides a simple interface to programming data that may cross memory
    region boundaries. To use it, create an instance and pass in the session object. Then call
    add_data() for each chunk of binary data you need to write. When all data is added, call the
    commit() method to write everything to memory. You may reuse a single MemoryLoader instance for
    multiple add-commit sequences.
    
    When programming across multiple regions, progress reports are combined so that only a
    one progress output is reported. Similarly, the programming performance report for each region
    is suppresed and a combined report is logged.
    
    Internally, MemoryBuilder instances are used to buffer data to be written to different types of memory.
    FlashBuilder is used to optimise programming within flash memory regions. RAM regions are programmed
    using the much simpler RamBuilder.
    """

    _session: "Session"
    _map: "MemoryMap"
    _progress: Optional[ProgressCallback]
    _builders: Dict["MemoryRegion", MemoryBuilder]
    _total_data_size: int
    _progress_offset: float
    _current_progress_fraction: float

    _chip_erase: Optional[bool]
    _smart_flash: Optional[bool]
    _trust_crc: Optional[bool]
    _keep_unwritten: Optional[bool]

    def __init__(self,
            session: "Session",
            progress: Optional[ProgressCallback] = None,
            chip_erase: Optional[bool] = None,
            smart_flash: Optional[bool] = None,
            trust_crc: Optional[bool] = None,
            keep_unwritten: Optional[bool] = None
        ):
        """! @brief Constructor.
        
        @param self
        @param session The session object.
        @param progress A progress report handler as a callable that takes a percentage completed.
            If not set or None, a default progress handler will be used unless the session option
            'hide_programming_progress' is set to True, in which case progress will be disabled.
        @param chip_erase Sets whether to use chip erase or sector erase. The value must be one of
            "auto", "sector", or "chip". "auto" means the fastest erase method should be used.
        @param smart_flash If set to True, the flash loader will attempt to not program pages whose
            contents are not going to change by scanning target flash memory. A value of False will
            force all pages to be erased and programmed.
        @param trust_crc Boolean indicating whether to use only the sector CRC32 to decide whether a
            sector already contains the data to be programmed. Use with caution, as CRC32 may return
            the same value for different content. Only applies if smart_flash is True.
        @param keep_unwritten Depending on the sector versus page size and the amount of data
            written, there may be ranges of flash that would be erased but not written with new
            data. This parameter sets whether the existing contents of those unwritten ranges will
            be read from memory and restored while programming.
        """
        self._session = session
        assert session.board
        target = session.board.target
        self._map = target.memory_map

        if progress is not None:
            self._progress = progress
        elif session.options.get('hide_programming_progress'):
            self._progress = None
        else:
            self._progress = print_progress()

        # We have to use a special sentinel object for chip_erase because None is a valid value.
        self._chip_erase = chip_erase if (chip_erase is not None) \
                            else self._session.options.get('chip_erase')
        self._smart_flash = smart_flash if (smart_flash is not None) \
                            else self._session.options.get('smart_flash')
        self._trust_crc = trust_crc if (trust_crc is not None) \
                            else self._session.options.get('fast_program')
        self._keep_unwritten = keep_unwritten if (keep_unwritten is not None) \
                            else self._session.options.get('keep_unwritten')
        
        self._reset_state()
    
    def _reset_state(self):
        """! @brief Clear all state variables. """
        # _builders is a dict that maps memory regions to either a FlashBuilder or, for writable memories,
        # a bytearray.
        self._builders = {}
        self._total_data_size = 0
        self._progress_offset = 0.0
        self._current_progress_fraction = 0.0
    
    def add_data(self, address, data):
        """! @brief Add a chunk of data to be programmed.
        
        The data may cross memory region boundaries, as long as the regions are contiguous.
        
        @param self
        @param address Integer address for where the first byte of _data_ should be written.
        @param data A list of byte values to be programmed at the given address.
        
        @return The MemoryLoader instance is returned, to allow chaining further add_data()
            calls or a call to commit().
        
        @exception ValueError Raised when the address is not within a flash memory region.
        @exception TargetSupportError Raised if the flash memory region does not have a valid Flash
            instance associated with it, which indicates that the target connect sequence did
            not run successfully.
        """
        while len(data):
            # Look up the memory region for this address.
            region = self._map.get_region_for_address(address)
            if region is None:
                raise ValueError("no memory region defined for address 0x%08x" % address)

            region_builder = self._builders.get(region, None)

            # Create the builder for this region if we don't already have one. This also verifies
            # that the region is of a type we can write to.
            if region_builder is None:
                if region.is_flash:
                    if region.flash is None:
                        raise exceptions.TargetSupportError(f"flash memory region at address {address:#010x} has no flash instance")
                    region_builder = region.flash.get_flash_builder()
                    region_builder.log_performance = False
                elif region.is_writable:
                    region_builder = RamBuilder(self._session, region)
                else:
                    raise ValueError(f"memory region at address {address:#010x} is not writable")
                
                # Save the new builder.
                assert region_builder is not None
                self._builders[region] = region_builder
        
            # Take as much data as is contained by this region.
            program_length = min(len(data), region.end - address + 1)
            assert program_length != 0

            # Add data to this region's builder.
            region_builder.add_data(address, data[:program_length])
            
            # Advance.
            data = data[program_length:]
            address += program_length
            self._total_data_size += program_length
        
        return self
    
    def commit(self):
        """! @brief Write all collected data to memory.
        
        This routine ensures that chip erase is only used once if either the auto mode or chip
        erase mode are used. As an example, if two regions are to be written to and True was
        passed to the constructor for chip_erase (or if the session option was set), then only
        the first region will actually use chip erase. The second region will be forced to use
        sector erase. This will not result in extra erasing, as sector erase always verifies whether
        the sectors are already erased. This will, of course, also work correctly if the flash
        algorithm for the first region doesn't actually erase the entire chip (all regions).
        
        After calling this method, the loader instance can be reused to program more data.
        """
        didChipErase = False
        perfList = []
        
        # Iterate over builders we've created and program the data.
        for builder in sorted(self._builders.values(), key=lambda v: v.region.start):
            # Determine this builder's portion of total progress.
            self._current_progress_fraction = builder.buffered_data_size / self._total_data_size
            
            # Program the data.
            chipErase = self._chip_erase if not didChipErase else "sector"
            perf = builder.program(chip_erase=chipErase,
                                    progress_cb=self._progress_cb,
                                    smart_flash=self._smart_flash,
                                    fast_verify=self._trust_crc,
                                    keep_unwritten=self._keep_unwritten)
            perfList.append(perf)
            didChipErase = True
            
            self._progress_offset += self._current_progress_fraction

        # Report programming statistics.
        self._log_performance(perfList)
        
        # Clear state to allow reuse.
        self._reset_state()
    
    def _log_performance(self, perf_list):
        """! @brief Log a report of programming performance numbers."""
        # Compute overall performance numbers.
        totalProgramTime = sum(perf.program_time for perf in perf_list)
        program_byte_count = sum(perf.total_byte_count for perf in perf_list)
        actual_program_byte_count = sum(perf.program_byte_count for perf in perf_list)
        actual_program_page_count = sum(perf.program_page_count for perf in perf_list)
        skipped_byte_count = sum(perf.skipped_byte_count for perf in perf_list)
        skipped_page_count = sum(perf.skipped_page_count for perf in perf_list)
        
        # Compute kbps while avoiding a potential zero-div error.
        if totalProgramTime == 0:
            kbps = 0
        else:
            kbps = (program_byte_count/1024) / totalProgramTime
        
        if any(perf.program_type == FlashBuilder.FLASH_CHIP_ERASE for perf in perf_list):
            LOG.info("Erased chip, programmed %d bytes (%s), skipped %d bytes (%s) at %.02f kB/s",
                actual_program_byte_count, get_page_count(actual_program_page_count),
                skipped_byte_count, get_page_count(skipped_page_count),
                kbps)
        else:
            erase_byte_count = sum(perf.erase_byte_count for perf in perf_list)
            erase_sector_count = sum(perf.erase_sector_count for perf in perf_list)

            LOG.info("Erased %d bytes (%s), programmed %d bytes (%s), skipped %d bytes (%s) at %.02f kB/s", 
                erase_byte_count, get_sector_count(erase_sector_count),
                actual_program_byte_count, get_page_count(actual_program_page_count),
                skipped_byte_count, get_page_count(skipped_page_count),
                kbps)
        
    def _progress_cb(self, amount):
        if self._progress is not None:
            self._progress((amount * self._current_progress_fraction) + self._progress_offset)
    
    @classmethod
    def program_binary_data(cls, session, address, data):
        """! @brief Helper routine to write a single chunk of data.
        
        The session options for chip_erase and trust_crc are used.
        
        @param cls
        @param session The session instance.
        @param address Start address of the data to program.
        @param data A list of byte values that will be programmed starting at _address_.
        """
        mgr = cls(session)
        mgr.add_data(address, data)
        mgr.commit()

# Define deprecated class name.
FlashLoader = MemoryLoader



#!/usr/bin/env python
# pyOCD debugger
# Copyright (c) 2006-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import os
import sys
from time import sleep
from random import randrange
import math
import logging

from pyocd.core.helpers import ConnectHelper
from pyocd.flash.file_programmer import FileProgrammer
from test_util import (
    get_session_options,
    get_test_binary_path,
    )

logging.basicConfig(level=logging.WARNING)

print("\n\n------ Test attaching to locked board ------")
for i in range(0, 10):
    with ConnectHelper.session_with_chosen_probe(**get_session_options()) as session:
        board = session.board
        flash = session.target.memory_map.get_boot_memory().flash
        # Erase and then reset - This locks Kinetis devices
        flash.init(flash.Operation.ERASE)
        flash.erase_all()
        flash.cleanup()
        board.target.reset()

print("\n\n------ Testing Attaching to board ------")
for i in range(0, 100):
    with ConnectHelper.session_with_chosen_probe(**get_session_options()) as session:
        board = session.board
        board.target.halt()
        sleep(0.01)
        board.target.resume()
        sleep(0.01)

print("\n\n------ Flashing new code ------")
with ConnectHelper.session_with_chosen_probe(**get_session_options()) as session:
    board = session.board
    binary_file = get_test_binary_path(board.test_binary)
    FileProgrammer(session).program(binary_file)

print("\n\n------ Testing Attaching to regular board ------")
for i in range(0, 10):
    with ConnectHelper.session_with_chosen_probe(**get_session_options()) as session:
        board = session.board
        board.target.reset_and_halt()
        board.target.halt()
        sleep(0.2)
        board.target.resume()
        sleep(0.2)



# Test user script.
#
# Provides stub implementations of all hooks.

def will_connect(board):
    """! @brief Pre-init hook for the board.
    @param self
    @param board A Board instance that is about to be initialized.
    @return Ignored.
    """
    pass

def did_connect(board):
    """! @brief Post-initialization hook for the board.
    @param self
    @param board A Board instance.
    @return Ignored.
    """
    pass

def will_init_target(target, init_sequence):
    """! @brief Hook to review and modify init call sequence prior to execution.
    @param self
    @param target A CoreSightTarget object about to be initialized.
    @param init_sequence The CallSequence that will be invoked. Because call sequences are
        mutable, this parameter can be modified before return to change the init calls.
    @return Ignored.
    """
    pass

def did_init_target(target):
    """! @brief Post-initialization hook.
    @param self
    @param target A CoreSightTarget.
    @return Ignored.
    """
    pass

def will_start_debug_core(core):
    """! @brief Hook to enable debug for the given core.
    @param self
    @param core A CortexM object about to be initialized.
    @retval True Do not perform the normal procedure to start core debug.
    @retval "False or None" Continue with normal behaviour.
    """
    pass

def did_start_debug_core(core):
    """! @brief Post-initialization hook.
    @param self
    @param core A CortexM object.
    @return Ignored.
    """
    pass

def will_stop_debug_core(core):
    """! @brief Pre-cleanup hook for the core.
    @param self
    @param core A CortexM object.
    @retval True Do not perform the normal procedure to disable core debug.
    @retval "False or None" Continue with normal behaviour.
    """
    pass

def did_stop_debug_core(core):
    """! @brief Post-cleanup hook for the core.
    @param self
    @param core A CortexM object.
    @return Ignored.
    """
    pass

def will_disconnect(target, resume):
    """! @brief Pre-disconnect hook.
    @param self
    @param target Either a CoreSightTarget or CortexM object.
    @param resume The value of the `disconnect_on_resume` option.
    @return Ignored.
    """
    pass

def did_disconnect(target, resume):
    """! @brief Post-disconnect hook.
    @param self
    @param target Either a CoreSightTarget or CortexM object.
    @param resume The value of the `disconnect_on_resume` option.
    @return Ignored."""
    pass

def will_reset(core, reset_type):
    """! @brief Pre-reset hook.
    @param self
    @param core A CortexM instance.
    @param reset_type One of the Target.ResetType enumerations.
    @retval True
    @retval "False or None"
    """
    pass

def did_reset(core, reset_type):
    """! @brief Post-reset hook.
    @param self
    @param core A CortexM instance.
    @param reset_type One of the Target.ResetType enumerations.
    @return Ignored.
    """
    pass

def set_reset_catch(core, reset_type):
    """! @brief Hook to prepare target for halting on reset.
    @param self
    @param core A CortexM instance.
    @param reset_type One of the Target.ResetType enumerations.
    @retval True
    @retval "False or None"
    """
    pass

def clear_reset_catch(core, reset_type):
    """! @brief Hook to clean up target after a reset and halt.
    @param self
    @param core A CortexM instance.
    @param reset_type
    @return Ignored.
    """
    pass

def mass_erase(target):
    """! @brief Hook to override mass erase.
    @param self
    @param target A CoreSightTarget object.
    @retval True Indicate that mass erase was performed by the hook.
    @retval "False or None" Mass erase was not overridden and the caller should proceed with the standard
        mass erase procedure.
    """
    pass

def trace_start(target, mode):
    """! @brief Hook to prepare for tracing the target.
    @param self
    @param target A CoreSightTarget object.
    @param mode The trace mode. Currently always 0 to indicate SWO.
    @return Ignored.
    """
    pass

def trace_stop(target, mode):
    """! @brief Hook to clean up after tracing the target.
    @param self
    @param target A CoreSightTarget object.
    @param mode The trace mode. Currently always 0 to indicate SWO.
    @return Ignored.
    """
    pass





# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import os
import sys
import traceback
import logging

from pyocd.core.helpers import ConnectHelper
from pyocd.probe.pydapaccess import DAPAccess
from pyocd.core.memory_map import MemoryType

from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    get_test_binary_path,
    TEST_DIR,
    )

TEST_USER_SCRIPT = os.path.join(TEST_DIR, "test_user_script.py")

class UserScriptTestResult(TestResult):
    def __init__(self):
        super(UserScriptTestResult, self).__init__(None, None, None)
        self.name = "user_script"

class UserScriptTest(Test):
    def __init__(self):
        super(UserScriptTest, self).__init__("User Script Test", user_script_test)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            result = UserScriptTestResult()
            result.passed = False
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def user_script_test(board_id):
    with ConnectHelper.session_with_chosen_probe(
            unique_id=board_id, user_script=TEST_USER_SCRIPT, **get_session_options()) as session:
        board = session.board
        target = session.target

        test_params = get_target_test_params(session)
        session.probe.set_clock(test_params['test_clock'])

        memory_map = target.get_memory_map()
        boot_region = memory_map.get_boot_memory()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)
        binary_file = get_test_binary_path(board.test_binary)
        
        test_pass_count = 0
        test_count = 0
        result = UserScriptTestResult()

        target.reset_and_halt()
        target.resume()
        target.halt()
        target.step()
        
        test_count += 1
        test_pass_count += 1
        
        print("\nTest Summary:")
        print("Pass count %i of %i tests" % (test_pass_count, test_count))
        if test_pass_count == test_count:
            print("USER SCRIPT TEST PASSED")
        else:
            print("USER SCRIPT TEST FAILED")

        target.reset()

        result.passed = test_count == test_pass_count
        return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD user script test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    DAPAccess.set_args(args.daparg)
    # Set to debug to print some of the decisions made while flashing
    session = ConnectHelper.session_with_chosen_probe(**get_session_options())
    test = UserScriptTest()
    result = [test.run(session.board)]




# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from time import sleep

from ..core import exceptions
from .component import CoreSightComponent
from .gpr import GPR
from .component_ids import COMPONENT_MAP
from ..utility.conversion import pairwise
from ..utility.mask import (bit_invert, align_down)
from ..utility.timeout import Timeout

LOG = logging.getLogger(__name__)

class CoreSightComponentID(object):
    """! @brief Reads and parses CoreSight architectural component ID registers.
    
    Reads the CIDR, PIDR, DEVID, and DEVARCH registers present at well known offsets
    in the memory map of all CoreSight components. The various fields from these
    registers are made available as attributes.
    """

    # CoreSight identification register offsets.
    DEVARCH = 0xfbc
    DEVID = 0xfc8
    DEVTYPE = 0xfcc
    PIDR4 = 0xfd0
    PIDR0 = 0xfe0
    CIDR0 = 0xff0
    IDR_END = 0x1000

    # Range of identification registers to read at once and offsets in results.
    #
    # To improve component identification performance, we read all of a components
    # CoreSight ID registers in a single read. Reading starts at the DEVARCH register.
    IDR_READ_START = DEVARCH
    IDR_READ_COUNT = (IDR_END - IDR_READ_START) // 4
    DEVARCH_OFFSET = (DEVARCH - IDR_READ_START) // 4
    DEVTYPE_OFFSET = (DEVTYPE - IDR_READ_START) // 4
    PIDR4_OFFSET = (PIDR4 - IDR_READ_START) // 4
    PIDR0_OFFSET = (PIDR0 - IDR_READ_START) // 4
    CIDR0_OFFSET = (CIDR0 - IDR_READ_START) // 4

    # Component ID register fields.
    CIDR_PREAMBLE_MASK = 0xffff0fff
    CIDR_PREAMBLE_VALUE = 0xb105000d

    CIDR_COMPONENT_CLASS_MASK = 0x0000f000
    CIDR_COMPONENT_CLASS_SHIFT = 12

    # Component classes.
    ROM_TABLE_CLASS = 0x1
    CORESIGHT_CLASS = 0x9

    # Peripheral ID register fields.
    PIDR_PART_MASK = 0x00000fff
    PIDR_DESIGNER_MASK = 0x0007f000 # JEP106 ID
    PIDR_DESIGNER_SHIFT = 12
    PIDR_REVISION_MASK = 0x00f00000
    PIDR_REVISION_SHIFT = 20
    PIDR_DESIGNER2_MASK = 0x0f00000000 # JEP106 continuation
    PIDR_DESIGNER2_SHIFT = 32

    # DEVARCH register fields.
    DEVARCH_ARCHITECT_MASK = 0x7ff
    DEVARCH_ARCHITECT_SHIFT = 21
    DEVARCH_PRESENT_MASK = (1<<20)
    DEVARCH_REVISION_MASK = 0x000f0000
    DEVARCH_REVISION_SHIFT = 16
    DEVARCH_ARCHID_MASK = 0xffff

    CLASS_0X9_ROM_TABLE_ARCHID = 0x0af7
    
    def __init__(self, parent_rom_table, ap, top_addr, power_id=None):
        self.parent_rom_table = parent_rom_table
        self.ap = ap
        self.address = top_addr
        self.top_address = top_addr
        self.power_id = power_id
        self.component_class = 0
        self.is_rom_table = False
        self.cidr = 0
        self.pidr = 0
        self.designer = 0
        self.part = 0
        self.devarch = 0
        self.archid = 0
        self.devtype = 0
        self.devid = [0, 0, 0]
        self.name = ''
        self.product_name = None
        self.factory = None
        self.valid = False

    def read_id_registers(self):
        """! @brief Read Component ID, Peripheral ID, and DEVID/DEVARCH registers."""
        # Read registers as a single block read for performance reasons.
        regs = self.ap.read_memory_block32(self.top_address + self.IDR_READ_START, self.IDR_READ_COUNT)
        self.cidr = self._extract_id_register_value(regs, self.CIDR0_OFFSET)
        self.pidr = (self._extract_id_register_value(regs, self.PIDR4_OFFSET) << 32) \
                    | self._extract_id_register_value(regs, self.PIDR0_OFFSET)

        # Check if the component has a valid CIDR value
        if (self.cidr & self.CIDR_PREAMBLE_MASK) != self.CIDR_PREAMBLE_VALUE:
            LOG.warning("Invalid coresight component, cidr=0x%x", self.cidr)
            return

        # Extract class.
        self.component_class = (self.cidr & self.CIDR_COMPONENT_CLASS_MASK) >> self.CIDR_COMPONENT_CLASS_SHIFT
        
        # Extract JEP106 designer ID.
        self.designer = ((self.pidr & self.PIDR_DESIGNER_MASK) >> self.PIDR_DESIGNER_SHIFT) \
                        | ((self.pidr & self.PIDR_DESIGNER2_MASK) >> (self.PIDR_DESIGNER2_SHIFT - 8))
        self.part = self.pidr & self.PIDR_PART_MASK
        
        # Handle Class 0x1 and Type 0x9 components.
        if self.component_class == self.ROM_TABLE_CLASS:
            # Class 0x1 ROM table.
            self.is_rom_table = True
        elif self.component_class == self.CORESIGHT_CLASS:
            # For CoreSight-class components, extract additional fields.
             self.devarch = regs[self.DEVARCH_OFFSET]
             self.devid = regs[1:4]
             self.devid.reverse()
             self.devtype = regs[self.DEVTYPE_OFFSET]
             
             if self.devarch & self.DEVARCH_PRESENT_MASK:
                 self.archid = self.devarch & self.DEVARCH_ARCHID_MASK

             # Identify a Class 0x9 ROM table.
             self.is_rom_table = (self.archid == self.CLASS_0X9_ROM_TABLE_ARCHID)
        
        # Determine component name.
        if self.is_rom_table:
            self.name = 'ROM'
            self.factory = ROMTable.create
        else:
            key = (self.designer, self.component_class, self.part, self.devtype, self.archid)
            info = COMPONENT_MAP.get(key, None)
            if info is None:
                # Try just the archid with no partno or devtype as backup.
                key = (self.designer, self.component_class, None, None, self.archid)
                info = COMPONENT_MAP.get(key, None)
            if info is not None:
                self.name = info.name
                self.product_name = info.product
                self.factory = info.factory
            else:
                self.name = '???'

        self.valid = True

    def _extract_id_register_value(self, regs, offset):
        result = 0
        for i in range(4):
            value = regs[offset + i]
            result |= (value & 0xff) << (i * 8)
        return result

    def __repr__(self):
        name = self.name
        if self.product_name:
            name += " " + self.product_name
        if not self.valid:
            return "<%08x:%s cidr=%x, pidr=%x, component invalid>" % (self.address, name, self.cidr, self.pidr)
        if self.power_id is not None:
            pwrid = " pwrid=%d" % self.power_id
        else:
            pwrid = ""
        if self.component_class == self.CORESIGHT_CLASS:
            return "<%08x:%s class=%d designer=%03x part=%03x devtype=%02x archid=%04x devid=%x:%x:%x%s>" % (
                self.address, name, self.component_class, self.designer, self.part,
                self.devtype, self.archid, self.devid[0], self.devid[1], self.devid[2], pwrid)
        else:
            return "<%08x:%s class=%d designer=%03x part=%03x%s>" % (
                self.address, name,self.component_class, self.designer, self.part, pwrid)


class ROMTable(CoreSightComponent):
    """! @brief CoreSight ROM table base class.
    
    This abstract class provides common functionality for ROM tables. Most importantly it has the
    static create() factory method.
    
    After a ROMTable is created, its init() method should be called. This will read and parse the
    table and any child ROM tables. For every component it finds in the table(s), it creates a
    CoreSightComponentID instance. The full collection of component IDs is available in the
    _components_ property. The for_each() method will execute a callable for all of the receiving
    ROM table and its children's components.
    
    Power domains controlled by Granular Power Requestor components are supported. They are
    automatically enabled as parsing proceeds so that components can be accessed to read their ID
    registers.
    """

    # 9 entries is enough entries to cover the standard Cortex-M4 ROM table for devices with ETM.
    ROM_TABLE_ENTRY_READ_COUNT = 9

    @staticmethod
    def create(memif, cmpid, addr=None, parent_table=None):
        """! @brief Factory method for creating ROM table components.
        
        This static method instantiates the appropriate subclass for the ROM table component
        described by the cmpid parameter.
        
        @param memif MemoryInterface used to access the ROM table.
        @param cmpid The CoreSightComponentID instance for this ROM table.
        @param addr Optional base address for this ROM table, if already known.
        @param parent_table Optional ROM table that pointed to this one.
        """
        assert cmpid is not None
        
        # Create appropriate ROM table class.
        if cmpid.component_class == CoreSightComponentID.ROM_TABLE_CLASS:
            return Class1ROMTable(memif, cmpid, addr, parent_table)
        elif cmpid.component_class == CoreSightComponentID.CORESIGHT_CLASS:
            return Class9ROMTable(memif, cmpid, addr, parent_table)
        else:
            raise exceptions.DebugError("unexpected ROM table device class (%s)" % cmpid)
    
    def __init__(self, ap, cmpid=None, addr=None, parent_table=None):
        """! @brief Constructor."""
        assert cmpid is not None
        assert cmpid.is_rom_table
        super(ROMTable, self).__init__(ap, cmpid, addr)
        assert self.address is not None
        if parent_table is not None:
            parent_table.add_child(self)
        self._depth = (self.parent.depth + 1) if self.parent else 0
        self._components = []
        self.name = 'ROM'
        self.gpr = None
    
    @property
    def depth(self):
        """! @brief Number of parent ROM tables."""
        return self._depth
    
    @property
    def components(self):
        """! @brief List of CoreSightComponentID instances for components found in this table.
        
        This property contains only the components for this ROM table, not any child tables.
        
        Child ROM tables will be represented in the list by ROMTable instances rather than
        CoreSightComponentID.
        """
        return self._components
    
    @property
    def depth_indent(self):
        """! @brief String of whitespace with a width corresponding to the table's depth.'"""
        return "  " * self._depth

    def init(self):
        """! @brief Read and parse the ROM table.
        
        As table entries for CoreSight components are read, a CoreSightComponentID instance will be
        created and the ID registers read. These ID objects are added to the _components_ property.
        If any child ROM tables are discovered, they will automatically be created and inited.
        """
        LOG.info("%s%s Class 0x%x ROM table #%d @ 0x%08x (designer=%03x part=%03x)",
            self.depth_indent, self.ap.short_description, self.cmpid.component_class, self.depth,
            self.address, self.cmpid.designer, self.cmpid.part)
        self._components = []

        self._read_table()
    
    def _read_table(self):
        raise NotImplementedError()

    def for_each(self, action, filter=None):
        """! @brief Apply an action to every component defined in the ROM table and child tables.
        
        This method iterates over every entry in the ROM table. For each entry it calls the
        filter function if provided. If the filter passes (returns True or was not provided) then
        the action function is called.
        
        The ROM table must have been initialized by calling init() prior to using this method.
        
        @param self This object.
        @param action Callable that accepts a single parameter, a CoreSightComponentID instance.
        @param filter Optional filter callable. Must accept a CoreSightComponentID instance and
            return a boolean indicating whether to perform the action (True applies action).
        """
        for component in self._components:
            # Recurse into child ROM tables.
            if isinstance(component, ROMTable):
                component.for_each(action, filter)
                continue
            
            # Skip component if the filter returns False.
            if filter is not None and not filter(component):
                continue
            
            # Perform the action.
            action(component)

class Class1ROMTable(ROMTable):
    """! @brief CoreSight Class 0x1 ROM table component and parser.
    
    An object of this class represents a CoreSight Class 0x1 ROM table. It supports reading the table
    and any child tables. For each entry in the table, a CoreSightComponentID object is created
    that further reads the component's CoreSight identification registers.
    
    Granular Power Requestor (GPR) components are supported to automatically enable power domains
    required to access components, as indicated by the component entry in the ROM table.
    """

    # Constants for Class 0x1 ROM tables.
    ROM_TABLE_ENTRY_PRESENT_MASK = 0x1

    # Mask for ROM table entry size. 1 if 32-bit entries.
    ROM_TABLE_32BIT_FORMAT_MASK = 0x2

    # ROM table entry power ID fields.
    ROM_TABLE_POWERIDVALID_MASK = 0x4
    ROM_TABLE_POWERID_MASK = 0x01f0
    ROM_TABLE_POWERID_SHIFT = 4

    # 2's complement offset to debug component from ROM table base address.
    ROM_TABLE_ADDR_OFFSET_NEG_MASK = 0x80000000
    ROM_TABLE_ADDR_OFFSET_MASK = 0xfffff000

    ROM_TABLE_MAX_ENTRIES = 960

    def _read_table(self):
        entryAddress = self.address
        foundEnd = False
        entriesRead = 0
        entryNumber = 0
        while not foundEnd and entriesRead < self.ROM_TABLE_MAX_ENTRIES:
            # Read several entries at a time for performance.
            readCount = min(self.ROM_TABLE_MAX_ENTRIES - entriesRead, self.ROM_TABLE_ENTRY_READ_COUNT)
            entries = self.ap.read_memory_block32(entryAddress, readCount)
            entriesRead += readCount

            for entry in entries:
                # Zero entry indicates the end of the table.
                if entry == 0:
                    foundEnd = True
                    break
                try:
                    self._handle_table_entry(entry, entryNumber)
                except exceptions.TransferError as err:
                    LOG.error("Error attempting to probe CoreSight component referenced by "
                            "ROM table entry #%d: %s", entryNumber, err,
                            exc_info=self.ap.dp.session.get_current().log_tracebacks)

                entryAddress += 4
                entryNumber += 1

    def _power_component(self, number, powerid, entry):
        if self.gpr is None:
            LOG.warning("ROM table entry #%d specifies power ID #%d, but no power requestor "
                "component has been seen; skipping component (entry=0x%08x)",
                number, powerid, entry)
            return False
    
        # Power up the domain.
        if not self.gpr.power_up_one(powerid):
            LOG.error("Failed to power up power domain #%d", powerid)
            return False
        else:
            LOG.info("Enabled power to power domain #%d", powerid)
            return True

    def _handle_table_entry(self, entry, number):
        # Nonzero entries can still be disabled, so check the present bit before handling.
        if (entry & self.ROM_TABLE_ENTRY_PRESENT_MASK) == 0:
            return
        # Verify the entry format is 32-bit.
        if (entry & self.ROM_TABLE_32BIT_FORMAT_MASK) == 0:
            return

        # Get the component's top 4k address.
        offset = entry & self.ROM_TABLE_ADDR_OFFSET_MASK
        if (entry & self.ROM_TABLE_ADDR_OFFSET_NEG_MASK) != 0:
            offset = ~bit_invert(offset)
        address = self.address + offset
        
        # Check power ID.
        if (entry & self.ROM_TABLE_POWERIDVALID_MASK) != 0:
            powerid = (entry & self.ROM_TABLE_POWERID_MASK) >> self.ROM_TABLE_POWERID_SHIFT
            
            # Attempt to power up this component. Skip this component if we the attempt fails.
            if not self._power_component(number, powerid, entry):
                return
        else:
            powerid = None

        # Create component instance.
        cmpid = CoreSightComponentID(self, self.ap, address, powerid)
        cmpid.read_id_registers()
        
        # Is this component a power requestor?
        if cmpid.factory == GPR.factory:
            # Create the GPR instance and stash it.
            self.gpr = cmpid.factory(self.ap, cmpid, None)
            self.gpr.init()

        LOG.info("%s[%d]%s", self.depth_indent, number, str(cmpid))

        # Recurse into child ROM tables.
        if cmpid.is_rom_table:
            cmp = ROMTable.create(self.ap, cmpid, address, parent_table=self)
            cmp.init()
        else:
            cmp = cmpid

        if cmp is not None:
            self.components.append(cmp)
    
class Class9ROMTable(ROMTable):
    """! @brief CoreSight Class 0x9 ROM table component and parser.
    
    Handles parsing of class 0x9 ROM tables as defined in ADIv6.
    
    In addition to GPR (Granular Power Requestor) components for power domain management, this class
    supports the optional power request functionality present in class 0x9 ROM tables.
    """

    # Constants for Class 0x9 ROM tables.
    ROM_TABLE_ENTRY_PRESENT_MASK = 0x3
    ROM_TABLE_ENTRY_POWERIDVALID_MASK = 0x4
    ROM_TABLE_ENTRY_POWERID_MASK = 0x01f0
    ROM_TABLE_ENTRY_POWERID_SHIFT = 4

    ROM_TABLE_ENTRY_NOT_PRESENT_FINAL = 0x0
    ROM_TABLE_ENTRY_NOT_PRESENT_NOT_FINAL = 0x2
    ROM_TABLE_ENTRY_PRESENT = 0x3

    ROM_TABLE_DBGPCRn = 0xa00
    ROM_TABLE_DBGPSRn = 0xa80
    ROM_TABLE_SYSPCRn = 0xb00
    ROM_TABLE_SYSPSRn = 0xb80
    ROM_TABLE_PRIDR0 = 0xc00
    ROM_TABLE_DBGRSTRR = 0xc10
    ROM_TABLE_DBGRSTAR = 0xc14
    ROM_TABLE_SYSRSTRR = 0xc18
    ROM_TABLE_SYSRSTAR = 0xc1c
    ROM_TABLE_AUTHSTATUS = 0xfb8

    ROM_TABLE_DBGPCRn_PRESENT_MASK = 0x00000001
    ROM_TABLE_DBGPCRn_PR_MASK = 0x00000002
    ROM_TABLE_DBGPSRn_PS_MASK = 0x00000003
    
    ROM_TABLE_DBGPSRn_PS_MAYBE_NOT_POWERED = 0x0
    ROM_TABLE_DBGPSRn_PS_IS_POWERED = 0x1
    ROM_TABLE_DBGPSRn_PS_MUST_REMAIN_POWERED = 0x3
    
    ROM_TABLE_PRIDR0_VERSION_MASK = 0x0000000f
    ROM_TABLE_PRIDR0_VERSION = 1 # Current version number of the power request functionality.

    ROM_TABLE_DEVID_CP_MASK = 0x00000040
    ROM_TABLE_DEVID_PRR_MASK = 0x00000020
    ROM_TABLE_DEVID_SYSMEM_MASK = 0x00000010
    ROM_TABLE_DEVID_FORMAT_MASK = 0x0000000f

    ROM_TABLE_FORMAT_32BIT = 0x0
    ROM_TABLE_FORMAT_64BIT = 0x1

    ROM_TABLE_MAX_ENTRIES = 512 # Maximum 32-bit entries.

    # 2's complement offset to debug component from ROM table base address.
    ROM_TABLE_ADDR_OFFSET_NEG_MASK = { 32: (1 << 31), 64: (1 << 63) }
    ROM_TABLE_ADDR_OFFSET_MASK = { 32: 0xfffff000, 64: 0xfffffffffffff000 }
    
    # 5 second timeout on power domain requests.
    POWER_REQUEST_TIMEOUT = 5.0
    
    def __init__(self, ap, cmpid=None, addr=None, parent_table=None):
        """! @brief Component constructor."""
        super(Class9ROMTable, self).__init__(ap, cmpid, addr, parent_table)
        
        self._pridr_version = None

        # Extract flags from DEVID.
        self._has_com_port = ((self.cmpid.devid[0] & self.ROM_TABLE_DEVID_CP_MASK) != 0)
        self._has_prr = ((self.cmpid.devid[0] & self.ROM_TABLE_DEVID_PRR_MASK) != 0)
        self._is_sysmem = ((self.cmpid.devid[0] & self.ROM_TABLE_DEVID_SYSMEM_MASK) != 0)
        is_64bit = ((self.cmpid.devid[0] & self.ROM_TABLE_DEVID_FORMAT_MASK) != 0)
        self._width = 64 if is_64bit else 32
        LOG.debug("cp=%d prr=%d sysmem=%d w=%d", self._has_com_port, self._has_prr, self._is_sysmem, self._width)
    
    @property
    def has_com_port(self):
        """! @brief Whether the ROM table includes COM Port functionality."""
        return self._has_com_port
    
    @property
    def has_prr(self):
        """! @brief Whether the ROM table includes power and reset requesting functionality."""
        return self._has_prr
    
    @property
    def is_sysmem(self):
        """! @brief Whether the ROM table is present in system memory."""
        return self._is_sysmem

    def _read_table(self):
        """! @brief Reads and parses the ROM table."""
        # Compute multipliers for 32- or 64-bit.
        entrySizeMultiplier = self._width // 32
        actualMaxEntries = self.ROM_TABLE_MAX_ENTRIES // entrySizeMultiplier
        # Ensure 64-bit format is read as pairs of 32-bit values.
        entryReadCount = align_down(self.ROM_TABLE_ENTRY_READ_COUNT, entrySizeMultiplier)
        
        entryAddress = self.address
        foundEnd = False
        entriesRead = 0
        entryNumber = 0
        
        while not foundEnd and entriesRead < actualMaxEntries:
            # Read several entries at a time for performance.
            readCount = min(actualMaxEntries - entriesRead, entryReadCount)
            entries = self.ap.read_memory_block32(entryAddress, readCount)
            entriesRead += readCount

            # For 64-bit entries, combine pairs of 32-bit values into single 64-bit value.
            if self._width == 64:
                entries = [(lo | (hi << 32)) for lo, hi in pairwise(entries)]
            
            for entry in entries:
                present = entry & self.ROM_TABLE_ENTRY_PRESENT_MASK
                
                # Zero entry indicates the end of the table.
                if present == self.ROM_TABLE_ENTRY_NOT_PRESENT_FINAL:
                    foundEnd = True
                    break
                elif present == self.ROM_TABLE_ENTRY_PRESENT:
                    try:
                        self._handle_table_entry(entry, entryNumber)
                    except exceptions.TransferError as err:
                        LOG.error("Error attempting to probe CoreSight component referenced by "
                                "ROM table entry #%d: %s", entryNumber, err,
                                exc_info=self.ap.dp.session.get_current().log_tracebacks)

                entryAddress += 4 * entrySizeMultiplier
                entryNumber += 1

    def _power_component(self, number, powerid, entry):
        """! @brief Enable power to a component defined by a ROM table entry."""
        if not self._has_prr:
            # Attempt GPR method of power domain enabling.
            return super(Class9ROMTable, self)._power_component(number, powerid, entry)
        
        # Check power request functionality version here so we can provide a nice warning message.
        if not self.check_power_request_version():
            LOG.warning("Class 0x9 ROM table #%d @ 0x%08x has unsupported version (%d) of power "
                        "request functionality, needed for entry #%d (entry=0x%08x). Skipping "
                        "component.", self.depth, self.address, self._pridr_version, number, entry)
            return False
        
        if not self.power_debug_domain(powerid):
            LOG.error("Failed to power up power domain #%d", powerid)
            return False
        else:
            LOG.info("Enabled power to power domain #%d", powerid)
            return True

    def _handle_table_entry(self, entry, number):
        """! @brief Parse one ROM table entry."""
        # Get the component's top 4k address.
        offset = entry & self.ROM_TABLE_ADDR_OFFSET_MASK[self._width]
        if (entry & self.ROM_TABLE_ADDR_OFFSET_NEG_MASK[self._width]) != 0:
            offset = ~bit_invert(offset, width=self._width)
        address = self.address + offset
        
        # Check power ID.
        if (entry & self.ROM_TABLE_ENTRY_POWERIDVALID_MASK) != 0:
            powerid = (entry & self.ROM_TABLE_ENTRY_POWERID_MASK) >> self.ROM_TABLE_ENTRY_POWERID_SHIFT
            
            # Attempt to power up this component. Skip this component if we the attempt fails.
            if not self._power_component(number, powerid, entry):
                return
        else:
            powerid = None

        # Create component instance.
        cmpid = CoreSightComponentID(self, self.ap, address, powerid)
        cmpid.read_id_registers()
        
        # Is this component a power requestor?
        if cmpid.factory == GPR.factory:
            # Create the GPR instance and stash it.
            self.gpr = cmpid.factory(self.ap, cmpid, None)
            self.gpr.init()

        LOG.info("%s[%d]%s", self.depth_indent, number, str(cmpid))

        # Recurse into child ROM tables.
        if cmpid.is_rom_table:
            cmp = ROMTable.create(self.ap, cmpid, address, parent_table=self)
            cmp.init()
        else:
            cmp = cmpid

        if cmp is not None:
            self._components.append(cmp)
    
    def check_power_request_version(self):
        """! @brief Verify the power request functionality version."""
        # Cache the PRIDR0 VERSION field the first time.
        if self._pridr_version is None:
            pridr = self.ap.read32(self.address + self.ROM_TABLE_PRIDR0)
            self._pridr_version = pridr & self.ROM_TABLE_PRIDR0_VERSION_MASK

        return self._pridr_version == self.ROM_TABLE_PRIDR0_VERSION

    def power_debug_domain(self, domain_id, enable=True):
        """! @brief Control power for a specified power domain managed by this ROM table."""
        # Compute register addresses for this power domain.
        dbgpcr_addr = self.address + self.ROM_TABLE_DBGPCRn + (4 * domain_id)
        dbgpsr_addr = self.address + self.ROM_TABLE_DBGPSRn + (4 * domain_id)
        
        # Check the domain request PRESENT bit.
        dbgpcr = self.ap.read32(dbgpcr_addr)
        if (dbgpcr & self.ROM_TABLE_DBGPCRn_PRESENT_MASK) == 0:
            LOG.warning("Power request functionality for power domain #%d is not present.",
                domain_id)
            return False
        
        # Check if the PR bit matches our request.
        pr = (dbgpcr & self.ROM_TABLE_DBGPCRn_PR_MASK) != 0
        if pr == enable:
            return True

        # If enabling, we need to check the power domain status. If the domain is in the "must
        # remain powered" state, then we have to wait for it to exit this state before re-enabling
        # power. See Figure D4-3 "Debug power request process", page D4-317, in the ADIv6
        # specification [IHI0074B].
        if enable:
            with Timeout(self.POWER_REQUEST_TIMEOUT) as time_out:
                while time_out.check():
                    power_status = self.ap.read32(dbgpsr_addr) & self.ROM_TABLE_DBGPSRn_PS_MASK
                    if power_status != self.ROM_TABLE_DBGPSRn_PS_MUST_REMAIN_POWERED:
                        break
                else:
                    LOG.warning("Power request handshake did not complete for power domain #%d.",
                        domain_id)
                    return False
        
        # Change power enable bit.
        if enable:
            dbgpcr |= self.ROM_TABLE_DBGPCRn_PR_MASK
        else:
            dbgpcr &= ~self.ROM_TABLE_DBGPCRn_PR_MASK
        self.ap.write32(dbgpcr_addr, dbgpcr)
        
        # Wait for status bits to update.
        with Timeout(self.POWER_REQUEST_TIMEOUT) as time_out:
            while time_out.check():
                power_status = self.ap.read32(dbgpsr_addr)
                if power_status == self.ROM_TABLE_DBGPSRn_PS_IS_POWERED:
                    # Sleep for 100 ms.
                    sleep(0.1)
                    break
                elif power_status == self.ROM_TABLE_DBGPSRn_PS_MUST_REMAIN_POWERED:
                    break
            else:
                LOG.warning("Power request handshake did not complete for power domain #%d.",
                    domain_id)
                return False

        # Successfully changed state of the power domain.
        return True



# pyOCD debugger
# Copyright (c) 2016,2018-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import six
from xml.etree import ElementTree
from itertools import groupby

from ..utility import conversion
from ..utility.mask import (align_up, round_up_div)
from ..core import exceptions
from ..core.target import Target
from ..core.memory_map import MemoryType
from . import signals

LOG = logging.getLogger(__name__)

MAP_XML_HEADER = b"""<?xml version="1.0"?>
<!DOCTYPE memory-map PUBLIC "+//IDN gnu.org//DTD GDB Memory Map V1.0//EN" "http://sourceware.org/gdb/gdb-memory-map.dtd">
"""

TARGET_XML_HEADER = b"""<?xml version="1.0"?>
<!DOCTYPE feature SYSTEM "gdb-target.dtd">
"""

## @brief Maps the fault code found in the IPSR to a GDB signal value.
FAULT = [
            signals.SIGSTOP,
            signals.SIGSTOP,    # Reset
            signals.SIGINT,     # NMI
            signals.SIGSEGV,    # HardFault
            signals.SIGSEGV,    # MemManage
            signals.SIGBUS,     # BusFault
            signals.SIGILL,     # UsageFault
                                                # The rest are not faults
         ]

## @brief Map from the memory type enums to gdb's memory region type names.
GDB_TYPE_MAP = {
    MemoryType.RAM: 'ram',
    MemoryType.ROM: 'rom',
    MemoryType.FLASH: 'flash',
    }

class GDBDebugContextFacade(object):
    """! @brief Provides GDB specific transformations to a DebugContext."""

    ## The order certain target features should appear in target XML.        
    REQUIRED_FEATURE_ORDER = ("org.gnu.gdb.arm.m-profile", "org.gnu.gdb.arm.vfp")

    def __init__(self, context):
        self._context = context
        
        # Note: Use the gdb 'maint print remote-registers' command to see it's view of the g/G commands.
        
        ## List of CoreRegisterInfos sorted by gdb_regnum, excluding any registers not communicated to gdb.
        #
        # This list is in the order expected by the g/G commands for reading/writing full register contexts.
        # It contains de-duplicated core registers with a valid GDB regnum, sorted by regnum.
        self._register_list = sorted(set(self._context.core.core_registers.iter_matching(
                lambda reg: reg.gdb_regnum is not None)), key=lambda v: v.gdb_regnum)
        
        ## List of internal register numbers corresponding to gdb registers.
        self._full_reg_num_list = [reg.index for reg in self._register_list]
        
        ## Map of gdb regnum to register info.
        self._gdb_regnum_map = {reg.gdb_regnum: reg for reg in self._register_list}
        
        ## String of XML target description for gdb.
        self._target_xml = self._build_target_xml()

    @property
    def context(self):
        return self._context

    def set_context(self, new_context):
        self._context = new_context

    def get_register_context(self):
        """! @brief Return hexadecimal dump of registers as expected by GDB.
        
        @exception CoreRegisterAccessError
        """
        LOG.debug("GDB getting register context")
        resp = b''
        try:
            vals = self._context.read_core_registers_raw(self._full_reg_num_list)
        except exceptions.CoreRegisterAccessError:
            vals = [None] * len(self._full_reg_num_list)
            
        for reg, reg_value in zip(self._register_list, vals):
            # Return x's to indicate unavailable register value.
            if reg_value is None:
                r = b"xx" * round_up_div(reg.bitsize, 8)
            else:
                r = six.b(conversion.uint_to_hex_le(reg_value, reg.bitsize))
            resp += r
            LOG.debug("GDB get_reg_context: %s = %s -> %s", reg.name,
                    "None" if (reg_value is None) else ("0x%08X" % reg_value), r)

        return resp

    def set_register_context(self, data):
        """! @brief Set registers from GDB hexadecimal string.
        
        @exception CoreRegisterAccessError
        """
        LOG.debug("GDB setting register context")
        reg_num_list = []
        reg_data_list = []
        offset = 0
        for reg in self._register_list:
            if offset >= len(data):
                break
            hex_byte_count = align_up(reg.bitsize // 4, 2)
            reg_data = data[offset:(offset + hex_byte_count)]
            reg_value = conversion.hex_le_to_uint(reg_data, reg.bitsize)
            offset += hex_byte_count
            reg_num_list.append(reg.index)
            reg_data_list.append(reg_value)
            LOG.debug("GDB reg: %s = 0x%X", reg.name, reg_value)
        self._context.write_core_registers_raw(reg_num_list, reg_data_list)

    def set_register(self, gdb_regnum, data):
        """! @brief Set single register from GDB hexadecimal string.
        
        @param self The object.
        @param gdb_regnum The regnum of register in target XML sent to GDB.
        @param data String of hex-encoded value for the register.
        
        @exception CoreRegisterAccessError
        """
        reg = self._gdb_regnum_map.get(gdb_regnum, None)
        if reg is not None:
            value = conversion.hex_le_to_uint(data, reg.bitsize)
            LOG.debug("GDB: write reg %s: 0x%X", reg.name, value)
            self._context.write_core_register_raw(reg.name, value)
        else:
            LOG.warning("GDB: attempt to set invalid register (regnum %d)", gdb_regnum)

    def gdb_get_register(self, gdb_regnum):
        """! @brief Set single core register.
        
        @param self The object.
        @param gdb_regnum The regnum of register in target XML sent to GDB.
        @return String of hex-encoded value for the register.
        
        @exception CoreRegisterAccessError
        """
        reg = self._gdb_regnum_map.get(gdb_regnum, None)
        if reg is None:
            return b''
        
        try:
            reg_value = self._context.read_core_register_raw(reg.name)
            resp = six.b(conversion.uint_to_hex_le(reg_value, reg.bitsize))
            LOG.debug("GDB reg: %s = 0x%X", reg.name, reg_value)
        except exceptions.CoreRegisterAccessError:
            # Return x's if the register read failed.
            resp = b"xx" * round_up_div(reg.bitsize, 8)
            LOG.debug("GDB reg: %s = <error reading>", reg.name)
        return resp

    def get_t_response(self, force_signal=None):
        """! @brief Returns a GDB T response string.
        
        This includes:
        - The signal encountered.
        - The current value of the important registers (sp, lr, pc).
        """
        if force_signal is not None:
            response = six.b('T' + conversion.byte_to_hex2(force_signal))
        else:
            response = six.b('T' + conversion.byte_to_hex2(self.get_signal_value()))

        # Append fp(r7), sp(r13), lr(r14), pc(r15)
        response += self._get_reg_index_value_pairs(['r7', 'sp', 'lr', 'pc'])

        return response

    def get_signal_value(self):
        if self._context.core.is_debug_trap():
            return signals.SIGTRAP

        # If not a fault then default to SIGSTOP
        signal = signals.SIGSTOP

        if self._context.core.is_vector_catch():
            fault = self._context.core.read_core_register('ipsr')
            try:
                signal = FAULT[fault]
            except IndexError:
                # Default to SIGSTOP as set above.
                pass

        LOG.debug("GDB lastSignal: %d", signal)
        return signal

    def _get_reg_index_value_pairs(self, reg_list):
        """! @brief Return register values as pairs.
        
        Returns a string like NN:MMMMMMMM;NN:MMMMMMMM;...
        for the T response string.  NN is the index of the
        register to follow MMMMMMMM is the value of the register.
        """
        result = b''
        try:
            reg_values = self._context.read_core_registers_raw(reg_list)
        except exceptions.CoreRegisterAccessError:
            reg_values = [None] * len(reg_list)

        for reg_name, reg_value in zip(reg_list, reg_values):
            reg = self._context.core.core_registers.by_name[reg_name]
            # Return x's if the register read failed.
            if reg_value is None:
                encoded_reg = "xx" * round_up_div(reg.bitsize, 8)
            else:
                encoded_reg = conversion.uint_to_hex_le(reg_value, reg.bitsize)
            result += six.b(conversion.byte_to_hex2(reg.gdb_regnum) + ':' + encoded_reg + ';')
        return result

    def get_memory_map_xml(self):
        """! @brief Generate GDB memory map XML.
        """
        root = ElementTree.Element('memory-map')
        for r in  self._context.core.memory_map:
            # Look up the region type name. Regions default to ram if gdb doesn't
            # have a concept of the region type.
            gdb_type = GDB_TYPE_MAP.get(r.type, 'ram')
            
            start = hex(r.start).rstrip("L")
            length = hex(r.length).rstrip("L")
            mem = ElementTree.SubElement(root, 'memory', type=gdb_type, start=start, length=length)
            if r.is_flash:
                prop = ElementTree.SubElement(mem, 'property', name='blocksize')
                prop.text = hex(r.blocksize).rstrip("L")
        return MAP_XML_HEADER + ElementTree.tostring(root)

    def _define_xpsr_control_fields(self, xml_feature):
        """! @brief Define XPSR and CONTROL register types with fields."""
        control = ElementTree.SubElement(xml_feature, 'flags', id="control", size="4")
        ElementTree.SubElement(control, "field", name="nPRIV", start="0", end="0", type="bool")
        ElementTree.SubElement(control, "field", name="SPSEL", start="1", end="1", type="bool")
        if self._context.core.has_fpu:
            ElementTree.SubElement(control, "field", name="FPCA", start="2", end="2", type="bool")
        if Target.SecurityState.SECURE in self._context.core.supported_security_states:
            ElementTree.SubElement(control, "field", name="SFPA", start="3", end="3", type="bool")

        apsr = ElementTree.SubElement(xml_feature, 'flags', id="apsr", size="4")
        ElementTree.SubElement(apsr, "field", name="N", start="31", end="31", type="bool")
        ElementTree.SubElement(apsr, "field", name="Z", start="30", end="30", type="bool")
        ElementTree.SubElement(apsr, "field", name="C", start="29", end="29", type="bool")
        ElementTree.SubElement(apsr, "field", name="V", start="28", end="28", type="bool")
        ElementTree.SubElement(apsr, "field", name="Q", start="27", end="27", type="bool")
        ElementTree.SubElement(apsr, "field", name="GE", start="16", end="19", type="int")

        ipsr = ElementTree.SubElement(xml_feature, 'struct', id="ipsr", size="4")
        ElementTree.SubElement(ipsr, "field", name="EXC", start="0", end="8", type="int")

        xpsr = ElementTree.SubElement(xml_feature, 'union', id="xpsr")
        ElementTree.SubElement(xpsr, "field", name="xpsr", type="uint32")
        ElementTree.SubElement(xpsr, "field", name="apsr", type="apsr")
        ElementTree.SubElement(xpsr, "field", name="ipsr", type="ipsr")

    def _build_target_xml(self):
        # Extract list of registers, group into gdb features.
        regs_sorted_by_feature = sorted(self._register_list, key=lambda r: r.gdb_feature) # Must sort for groupby().
        regs_by_feature = {k: list(g) for k, g in groupby(regs_sorted_by_feature, key=lambda r: r.gdb_feature)}
        unordered_features = list(regs_by_feature.keys())
        features = []
        
        # Get a list of gdb features with some features having a determined order.
        for feature_name in self.REQUIRED_FEATURE_ORDER:
            if feature_name in unordered_features:
                features.append(feature_name)
                unordered_features.remove(feature_name)
        # Add any remaining features at the end of the feature list.
        features += unordered_features
        
        use_xpsr_control_fields = self._context.session.options.get('xpsr_control_fields')
        
        xml_root = ElementTree.Element('target')
        
        for feature_name in features:
            regs = regs_by_feature[feature_name]
        
            xml_feature = ElementTree.SubElement(xml_root, "feature", name=feature_name)

            # Special case for XPSR and CONTROL bitfield presentation.
            if (feature_name == "org.gnu.gdb.arm.m-profile") and use_xpsr_control_fields:
                self._define_xpsr_control_fields(xml_feature)
            
            # Add XML for the registers in this feature.
            for reg in regs:
                if use_xpsr_control_fields and (reg.name in ('xpsr', 'control')):
                    reg_type = reg.name
                else:
                    reg_type = reg.gdb_type
                ElementTree.SubElement(xml_feature, 'reg', name=reg.name, bitsize=str(reg.bitsize),
                        type=reg_type, group=reg.group, regnum=str(reg.gdb_regnum))

        return TARGET_XML_HEADER + ElementTree.tostring(xml_root)

    def get_target_xml(self):
        return self._target_xml

    def flush(self):
        self._context.core.flush()





# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import argparse
import os
import sys
from time import (sleep, time)
from random import randrange
import math
import struct
import traceback
import argparse
import logging
from itertools import (chain, repeat)

from pyocd.core.helpers import ConnectHelper
from pyocd.flash.file_programmer import FileProgrammer
from pyocd.probe.pydapaccess import DAPAccess
from pyocd.utility.conversion import float32_to_u32
from pyocd.utility.mask import same
from pyocd.utility.compatibility import to_str_safe
from pyocd.core.memory_map import MemoryType

from test_util import (
    Test,
    TestResult,
    get_session_options,
    get_target_test_params,
    run_in_parallel,
    )

# Test configuration values.        
TEST_MAX_LENGTH = 1 * 1024 * 1024
TEST_THREAD_COUNT = 8
TEST_SUBCHUNK_COUNT = 2 # Number of reads/writes per thread.

def ncycles(iterable, n):
    return chain.from_iterable(repeat(tuple(iterable), n))

class ConcurrencyTestResult(TestResult):
    def __init__(self):
        super(ConcurrencyTestResult, self).__init__(None, None, None)
        self.name = "concurrency"

class ConcurrencyTest(Test):
    def __init__(self):
        super(ConcurrencyTest, self).__init__("Concurrency Test", concurrency_test)

    def run(self, board):
        try:
            result = self.test_function(board.unique_id)
        except Exception as e:
            result = ConcurrencyTestResult()
            result.passed = False
            print("Exception %s when testing board %s" % (e, board.unique_id))
            traceback.print_exc(file=sys.stdout)
        result.board = board
        result.test = self
        return result

def concurrency_test(board_id):
    with ConnectHelper.session_with_chosen_probe(unique_id=board_id, **get_session_options()) as session:
        board = session.board
        target = session.target

        test_params = get_target_test_params(session)
        session.probe.set_clock(test_params['test_clock'])

        memory_map = target.get_memory_map()
        boot_region = memory_map.get_boot_memory()
        ram_region = memory_map.get_default_region_of_type(MemoryType.RAM)

        test_pass_count = 0
        test_count = 0
        result = ConcurrencyTestResult()
        
        target.reset_and_halt()
        
        # Prepare TEST_THREAD_COUNT regions of RAM with patterns
        data_len = min(TEST_MAX_LENGTH, ram_region.length)
        chunk_len = data_len // TEST_THREAD_COUNT
        subchunk_len = chunk_len // TEST_SUBCHUNK_COUNT
        
        chunk_data = []
        for i in range(TEST_THREAD_COUNT):
            chunk_data.append([(i + j) % 256 for j in range(chunk_len)])
        
        def write_chunk_data(core, i):
            start = ram_region.start + chunk_len * i
            for j in range(TEST_SUBCHUNK_COUNT):
                offset = subchunk_len * j
                addr = start + offset
                end = addr + subchunk_len - 1
                print("Writing region %i:%i from %#010x to %#010x via %s" % (i, j, addr, end, core.ap))
                core.write_memory_block8(addr, chunk_data[i][offset:offset + subchunk_len])
                print("Finished writing region %i:%i" % (i, j))

        def read_chunk_data(core, i):
            start = ram_region.start + chunk_len * i
            for j in range(TEST_SUBCHUNK_COUNT):
                offset = subchunk_len * j
                addr = start + offset
                end = addr + subchunk_len - 1
                print("Reading region %i:%i from %#010x to %#010x via %s" % (i, j, addr, end, core.ap))
                data = core.read_memory_block8(addr, subchunk_len)
                chunk_read_data[i].extend(data)
                print("Finished reading region %i:%i" % (i, j))
        
        # Test with a single core/AP.
        print("\n------ Test 1: Concurrent memory accesses, single core ------")
        
        core = target.cores[0]

        # Write chunk patterns concurrently.
        print("Writing %i regions to RAM" % TEST_THREAD_COUNT)
        run_in_parallel(write_chunk_data, [[core, i] for i in range(TEST_THREAD_COUNT)])
        
        print("Reading %i regions to RAM" % TEST_THREAD_COUNT)
        chunk_read_data = [list() for i in range(TEST_THREAD_COUNT)]
        run_in_parallel(read_chunk_data, [[core, i] for i in range(TEST_THREAD_COUNT)])
        
        print("Comparing data")
        
        for i in range(TEST_THREAD_COUNT):
            test_count += 1
            if same(chunk_read_data[i], chunk_data[i]):
                test_pass_count += 1
                print("Region %i PASSED" % i)
            else:
                print("Region %i FAILED" % i)
        
        # Test with a multiple cores/APs.
        # Disabled until cores each have their own memory map, the regions accessible to each
        # core can be identified.
        if False: # len(target.cores) > 1:
            print("\n------ Test 2: Concurrent memory accesses, multiple cores ------")
            
            cycle_count = ((len(target.cores) + TEST_THREAD_COUNT - 1) // TEST_THREAD_COUNT * TEST_THREAD_COUNT)
            repeat_cores = ncycles(iter(target.cores), cycle_count)
            thread_args = []
            for i in range(TEST_THREAD_COUNT):
                thread_args.append((target.cores[next(repeat_cores)], i))

            # Write chunk patterns concurrently.
            print("Writing %i regions to RAM" % TEST_THREAD_COUNT)
            run_in_parallel(write_chunk_data, thread_args)
        
            print("Reading %i regions to RAM" % TEST_THREAD_COUNT)
            chunk_read_data = [list() for i in range(TEST_THREAD_COUNT)]
            run_in_parallel(read_chunk_data, thread_args)
        
            print("Comparing data")
        
            for i in range(TEST_THREAD_COUNT):
                test_count += 1
                if same(chunk_read_data[i], chunk_data[i]):
                    test_pass_count += 1
                    print("Region %i PASSED" % i)
                else:
                    print("Region %i FAILED" % i)

        # --- end ---
        print("\nTest Summary:")
        print("Pass count %i of %i tests" % (test_pass_count, test_count))
        if test_pass_count == test_count:
            print("CONCURRENCY TEST PASSED")
        else:
            print("CONCURRENCY TEST FAILED")

        target.reset()

        result.passed = test_count == test_pass_count
        return result

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='pyOCD concurrency test')
    parser.add_argument('-d', '--debug', action="store_true", help='Enable debug logging')
    parser.add_argument("-da", "--daparg", dest="daparg", nargs='+', help="Send setting to DAPAccess layer.")
    args = parser.parse_args()
    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    DAPAccess.set_args(args.daparg)
    # Set to debug to print some of the decisions made while flashing
    session = ConnectHelper.session_with_chosen_probe(**get_session_options())
    test = ConcurrencyTest()
    result = [test.run(session.board)]




# pyOCD debugger
# Copyright (c) 2015-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import traceback
import atexit

from ..core import (session, exceptions)

LOG = logging.getLogger(__name__)

class ToolExitException(Exception):
    """! @brief Special exception indicating the tool should exit.
    
    This exception is only raised by the `exit` command.
    """
    pass

class PyocdRepl(object):
    """! @brief Read-Eval-Print-Loop for pyOCD commander."""
    
    PROMPT = 'pyocd> '
    
    PYOCD_HISTORY_ENV_VAR = 'PYOCD_HISTORY'
    PYOCD_HISTORY_LENGTH_ENV_VAR = 'PYOCD_HISTORY_LENGTH'
    DEFAULT_HISTORY_FILE = ".pyocd_history"

    def __init__(self, command_context):
        self.context = command_context
        
        # Attempt to import readline. If we import readline from the module level, it may be imported
        # when initing non-interactive subcommands such as gdbserver, and it causes the process to be
        # stopped if started in the background (& suffix in sh).
        try:
            import readline

            # Enable readline history.
            self._history_path = os.environ.get(self.PYOCD_HISTORY_ENV_VAR,
                    os.path.join(os.path.expanduser("~"), self.DEFAULT_HISTORY_FILE))
        
            # Read command history and set history length.
            try:
                readline.read_history_file(self._history_path)
            
                history_len = int(os.environ.get(self.PYOCD_HISTORY_LENGTH_ENV_VAR,
                        session.Session.get_current().options.get('commander.history_length')))
                readline.set_history_length(history_len)
            except (NameError, IOError) as err:
                pass

            # Install exit handler to write out the command history.
            try:
                atexit.register(readline.write_history_file, self._history_path)
            except (NameError, IOError) as err:
                pass
        except ImportError:
            pass

    def run(self):
        """! @brief Runs the REPL loop until EOF is encountered."""
        try:
            while True:
                try:
                    line = input(self.PROMPT)
                    self.run_one_command(line)
                except KeyboardInterrupt:
                    print()
        except EOFError:
            # Print a newline when we get a Ctrl-D on a Posix system.
            # Windows exits with a Ctrl-Z+Return, so there is no need for this.
            if os.name != "nt":
                print()
        except ToolExitException:
            pass
    
    def run_one_command(self, line):
        """! @brief Execute a single command line and handle exceptions."""
        try:
            line = line.strip()
            if line:
                self.context.process_command_line(line)
        except KeyboardInterrupt:
            print()
        except ValueError:
            print("Error: invalid argument")
            if session.Session.get_current().log_tracebacks:
                traceback.print_exc()
        except exceptions.TransferError as e:
            print("Transfer failed:", e)
            if session.Session.get_current().log_tracebacks:
                traceback.print_exc()
        except exceptions.CommandError as e:
            print("Error:", e)
        except ToolExitException:
            # Catch and reraise this exception so it isn't caught by the catchall below.
            raise
        except Exception as e:
            # Catch most other exceptions so they don't cause the REPL to exit.
            print("Error:", e)
            if session.Session.get_current().log_tracebacks:
                traceback.print_exc()



#!/usr/bin/env python

# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2020 Cypress Semiconductor Corporation
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import logging
import argparse
import colorama
import fnmatch
from typing import (Any, Optional, Sequence)

from . import __version__
from .core.session import Session
from .core import exceptions
from .probe.pydapaccess import DAPAccess
from .core import options
from .subcommands.base import SubcommandBase
from .subcommands.commander_cmd import CommanderSubcommand
from .subcommands.erase_cmd import EraseSubcommand
from .subcommands.gdbserver_cmd import GdbserverSubcommand
from .subcommands.json_cmd import JsonSubcommand
from .subcommands.list_cmd import ListSubcommand
from .subcommands.load_cmd import LoadSubcommand
from .subcommands.pack_cmd import PackSubcommand
from .subcommands.reset_cmd import ResetSubcommand
from .subcommands.server_cmd import ServerSubcommand

## @brief Default log format for all subcommands.
LOG_FORMAT = "%(relativeCreated)07d:%(levelname)s:%(module)s:%(message)s"

## @brief Logger for this module.
LOG = logging.getLogger("pyocd.tool")

class PyOCDTool(SubcommandBase):
    """! @brief Main class for the pyocd tool and subcommands.
    """
    
    HELP = "PyOCD debug tools for Arm Cortex devices"
    
    ## List of subcommand classes.
    SUBCOMMANDS = [
        CommanderSubcommand,
        EraseSubcommand,
        LoadSubcommand,
        GdbserverSubcommand,
        JsonSubcommand,
        ListSubcommand,
        PackSubcommand,
        ResetSubcommand,
        ServerSubcommand,
        ]
    
    ## @brief Logging level names.
    LOG_LEVEL_NAMES = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
            }

    def __init__(self):
        # Start with an empty namespace.
        super().__init__(argparse.Namespace())
        self._parser = self.build_parser()

    def build_parser(self) -> argparse.ArgumentParser:
        """! @brief Construct the command line parser with all subcommands and options."""
        # Create top level argument parser.
        parser = argparse.ArgumentParser(description=self.HELP)
        parser.set_defaults(command_class=self, quiet=0, verbose=0, log_level=[])

        parser.add_argument('-V', '--version', action='version', version=__version__)
        parser.add_argument('--help-options', action='store_true',
            help="Display available session options.")
            
        self.add_subcommands(parser)

        return parser

    def _setup_logging(self) -> None:
        """! @brief Configure the logging module.
        
        The quiet and verbose argument counts are used to set the log verbosity level.
        
        Log level for specific loggers are also configured here.
        """
        level = max(1, self._args.command_class.DEFAULT_LOG_LEVEL + self._get_log_level_delta())
        logging.basicConfig(level=level, format=LOG_FORMAT)
        
        # Handle settings for individual loggers from --log-level arguments.
        for logger_setting in self._args.log_level:
            try:
                loggers, level_name = logger_setting.split('=')[:2]
                level = self.LOG_LEVEL_NAMES[level_name.strip().lower()]
                for logger_pattern in loggers.split(','):
                    matching_loggers = fnmatch.filter(logging.root.manager.loggerDict.keys(), logger_pattern.strip()) # type:ignore
                    LOG.debug('setting log level %s for %s', level_name, matching_loggers)
                    for logger in matching_loggers:
                        log = logging.getLogger(logger)
                        log.setLevel(level)
                        log.disabled = False
            except (ValueError, KeyError):
                raise exceptions.CommandError(f"invalid --log-level argument '{logger_setting}'")
            except AttributeError:
                LOG.warning("Failed to set logger levels; logging module may have changed.")
                break

    def invoke(self) -> int:
        """! @brief Show help when pyocd is run with no subcommand."""
        if self._args.help_options:
            self.show_options_help()
        else:
            self._parser.print_help()
        return 0
    
    def __call__(self, *args: Any, **kwds: Any) -> "PyOCDTool":
        """! @brief Hack to allow the root command object instance to be used as default command class."""
        return self

    def run(self, args: Optional[Sequence[str]] = None) -> int:
        """! @brief Main entry point for command line processing."""
        try:
            self._args = self._parser.parse_args(args)
            
            self._setup_logging()
            
            # Pass any options to DAPAccess.
            if hasattr(self._args, 'daparg'):
                DAPAccess.set_args(self._args.daparg)
            
            # Create an instance of the subcommand and invoke it.
            cmd = self._args.command_class(self._args)
            status = cmd.invoke()

            # Successful exit.
            return status
        except KeyboardInterrupt:
            return 0
        except (exceptions.Error, ValueError, IndexError) as e:
            LOG.critical(e, exc_info=Session.get_current().log_tracebacks)
            return 1
        except Exception as e:
            LOG.critical("uncaught exception: %s", e, exc_info=Session.get_current().log_tracebacks)
            return 1
    
    def show_options_help(self) -> None:
        """! @brief Display help for session options."""
        for info_name in sorted(options.OPTIONS_INFO.keys()):
            info = options.OPTIONS_INFO[info_name]
            if isinstance(info.type, tuple):
                typename = ", ".join(t.__name__ for t in info.type)
            else:
                typename = info.type.__name__
            print((colorama.Fore.CYAN + colorama.Style.BRIGHT + "{name}" + colorama.Style.RESET_ALL  # type:ignore
                + colorama.Fore.GREEN + " ({typename})" + colorama.Style.RESET_ALL
                + " {help}").format(
                name=info.name, typename=typename, help=info.help))

def main():
    sys.exit(PyOCDTool().run())

if __name__ == '__main__':
    main()



# pyOCD debugger
# Copyright (c) 2015-2017 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.target import Target

class Breakpoint(object):
    def __init__(self, provider):
        self.type = Target.BreakpointType.HW
        self.enabled = False
        self.addr = 0
        self.original_instr = 0
        self.provider = provider

    def __repr__(self):
        return "<%s@0x%08x type=%s addr=0x%08x>" % (self.__class__.__name__, id(self), self.type.name, self.addr)

class BreakpointProvider(object):
    """! @brief Abstract base class for breakpoint providers."""
    def init(self):
        raise NotImplementedError()

    def bp_type(self):
        return 0

    @property
    def do_filter_memory(self):
        return False

    @property
    def available_breakpoints(self):
        raise NotImplementedError()

    def find_breakpoint(self, addr):
        raise NotImplementedError()

    def set_breakpoint(self, addr):
        raise NotImplementedError()

    def remove_breakpoint(self, bp):
        raise NotImplementedError()

    def filter_memory(self, addr, size, data):
        return data

    def flush(self):
        pass






# pyOCD debugger
# Copyright (c) 2016-2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from intervaltree import (Interval, IntervalTree)
import logging

from ..utility import conversion
from .metrics import CacheMetrics
from ..core.exceptions import TransferFaultError

LOG = logging.getLogger(__name__)

class MemoryCache(object):
    """! @brief Memory cache.
    
    Maintains a cache of target memory. The constructor is passed a backing DebugContext object that
    will be used to fill the cache.
    
    The cache is invalidated whenever the target has run since the last cache operation (based on run
    tokens). If the target is currently running, all accesses cause the cache to be invalidated.
    
    The target's memory map is referenced. All memory accesses must be fully contained within a single
    memory region, or a TransferFaultError will be raised. However, if an access is outside of all regions,
    the access is passed to the underlying context unmodified. When an access is within a region, that
    region's cacheability flag is honoured.
    """
    
    def __init__(self, context, core):
        self._context = context
        self._core = core
        self._run_token = -1
        self._reset_cache()

    def _reset_cache(self):
        self._cache = IntervalTree()
        self._metrics = CacheMetrics()

    def _check_cache(self):
        """! @brief Invalidates the cache if appropriate."""
        if self._core.is_running():
            LOG.debug("core is running; invalidating cache")
            self._reset_cache()
        elif self._run_token != self._core.run_token:
            self._dump_metrics()
            LOG.debug("out of date run token; invalidating cache")
            self._reset_cache()
            self._run_token = self._core.run_token

    def _get_ranges(self, addr, count):
        """! @brief Splits a memory address range into cached and uncached subranges.
        @return Returns a 2-tuple with the first element being a set of Interval objects for each
          of the cached subranges. The second element is a set of Interval objects for each of the
          non-cached subranges.
        """
        cached = self._cache.overlap(addr, addr + count)
        uncached = {Interval(addr, addr + count)}
        for cachedIv in cached:
            newUncachedSet = set()
            for uncachedIv in uncached:

                # No overlap.
                if cachedIv.end < uncachedIv.begin or cachedIv.begin > uncachedIv.end:
                    newUncachedSet.add(uncachedIv)
                    continue

                # Begin segment.
                if cachedIv.begin - uncachedIv.begin > 0:
                    newUncachedSet.add(Interval(uncachedIv.begin, cachedIv.begin))

                # End segment.
                if uncachedIv.end - cachedIv.end > 0:
                    newUncachedSet.add(Interval(cachedIv.end, uncachedIv.end))
            uncached = newUncachedSet
        return cached, uncached

    def _read_uncached(self, uncached):
        """! "@brief Reads uncached memory ranges and updates the cache.
        @return A list of Interval objects is returned. Each Interval has its @a data attribute set
          to a bytearray of the data read from target memory.
        """
        uncachedData = []
        for uncachedIv in uncached:
            data = self._context.read_memory_block8(uncachedIv.begin, uncachedIv.end - uncachedIv.begin)
            iv = Interval(uncachedIv.begin, uncachedIv.end, bytearray(data))
            self._cache.add(iv) # TODO merge contiguous cached intervals
            uncachedData.append(iv)
        return uncachedData

    def _update_metrics(self, cached, uncached, addr, size):
        cachedSize = 0
        for iv in cached:
            begin = iv.begin
            end = iv.end
            if iv.begin < addr:
                begin = addr
            if iv.end > addr + size:
                end = addr + size
            cachedSize += end - begin

        uncachedSize = sum((iv.end - iv.begin) for iv in uncached)

        self._metrics.reads += 1
        self._metrics.hits += cachedSize
        self._metrics.misses += uncachedSize

    def _dump_metrics(self):
        if self._metrics.total > 0:
            LOG.debug("%d reads, %d bytes [%d%% hits, %d bytes]; %d bytes written",
                self._metrics.reads, self._metrics.total, self._metrics.percent_hit,
                self._metrics.hits, self._metrics.writes)
        else:
            LOG.debug("no reads")

    def _read(self, addr, size):
        """! @brief Performs a cached read operation of an address range.
        @return A list of Interval objects sorted by address.
        """
        # Get the cached and uncached subranges of the requested read.
        cached, uncached = self._get_ranges(addr, size)
        self._update_metrics(cached, uncached, addr, size)

        # Read any uncached ranges.
        uncachedData = self._read_uncached(uncached)

        # Merged cached with data we just read
        combined = list(cached) + uncachedData
        combined.sort(key=lambda x: x.begin)
        return combined

    def _merge_data(self, combined, addr, size):
        """! @brief Extracts data from the intersection of an address range across a list of interval objects.
        
        The range represented by @a addr and @a size are assumed to overlap the intervals. The first
        and last interval in the list may have ragged edges not fully contained in the address range, in
        which case the correct slice of those intervals is extracted.
        
        @param self
        @param combined List of Interval objects forming a contiguous range. The @a data attribute of
          each interval must be a bytearray.
        @param addr Start address. Must be within the range of the first interval.
        @param size Number of bytes. (@a addr + @a size) must be within the range of the last interval.
        @return A single bytearray object with all data from the intervals that intersects the address
          range.
        """
        result = bytearray()
        resultAppend = bytearray()

        # Check for fully contained subrange.
        if len(combined) and combined[0].begin < addr and combined[0].end > addr + size:
            offset = addr - combined[0].begin
            endOffset = offset + size
            result = combined[0].data[offset:endOffset]
            return result
        
        # Take slice of leading ragged edge.
        if len(combined) and combined[0].begin < addr:
            offset = addr - combined[0].begin
            result += combined[0].data[offset:]
            combined = combined[1:]
        # Take slice of trailing ragged edge.
        if len(combined) and combined[-1].end > addr + size:
            offset = addr + size - combined[-1].begin
            resultAppend = combined[-1].data[:offset]
            combined = combined[:-1]

        # Merge.
        for iv in combined:
            result += iv.data
        result += resultAppend

        return result

    def _update_contiguous(self, cached, addr, value):
        size = len(value)
        end = addr + size
        leadBegin = addr
        leadData = bytearray()
        trailData = bytearray()
        trailEnd = end

        if cached[0].begin < addr and cached[0].end > addr:
            offset = addr - cached[0].begin
            leadData = cached[0].data[:offset]
            leadBegin = cached[0].begin
        if cached[-1].begin < end and cached[-1].end > end:
            offset = end - cached[-1].begin
            trailData = cached[-1].data[offset:]
            trailEnd = cached[-1].end

        self._cache.remove_overlap(addr, end)

        data = leadData + value + trailData
        self._cache.addi(leadBegin, trailEnd, data)

    def _check_regions(self, addr, count):
        """! @return A bool indicating whether the given address range is fully contained within
              one known memory region, and that region is cacheable.
        @exception TransferFaultError Raised if the access is not entirely contained within a single region.
        """
        regions = self._core.memory_map.get_intersecting_regions(addr, length=count)

        # If no regions matched, then allow an uncached operation.
        if len(regions) == 0:
            return False

        # Raise if not fully contained within one region.
        if len(regions) > 1 or not regions[0].contains_range(addr, length=count):
            raise TransferFaultError("individual memory accesses must not cross memory region boundaries")

        # Otherwise return whether the region is cacheable.
        return regions[0].is_cacheable

    def read_memory(self, addr, transfer_size=32, now=True):
        # TODO use more optimal underlying read_memory calls
        if transfer_size == 8:
            data = self.read_memory_block8(addr, 1)[0]
        else:
            data = conversion.byte_list_to_nbit_le_list(self.read_memory_block8(addr, transfer_size // 8),
                    transfer_size)[0]

        if now:
            return data
        else:
            def read_cb():
                return data
            return read_cb

    def read_memory_block8(self, addr, size):
        if size <= 0:
            return []

        self._check_cache()

        # Validate memory regions.
        if not self._check_regions(addr, size):
            LOG.debug("range [%x:%x] is not cacheable", addr, addr+size)
            return self._context.read_memory_block8(addr, size)

        # Get the cached and uncached subranges of the requested read.
        combined = self._read(addr, size)

        # Extract data out of combined intervals.
        result = list(self._merge_data(combined, addr, size))
        assert len(result) == size, "result size ({}) != requested size ({})".format(len(result), size)
        return result

    def read_memory_block32(self, addr, size):
        return conversion.byte_list_to_u32le_list(self.read_memory_block8(addr, size*4))

    def write_memory(self, addr, value, transfer_size=32):
        if transfer_size == 8:
            return self.write_memory_block8(addr, [value])
        else:
            return self.write_memory_block8(addr, conversion.nbit_le_list_to_byte_list([value], transfer_size))

    def write_memory_block8(self, addr, value):
        if len(value) <= 0:
            return

        self._check_cache()

        # Validate memory regions.
        cacheable = self._check_regions(addr, len(value))

        # Write to the target first, so if it fails we don't update the cache.
        result = self._context.write_memory_block8(addr, value)

        if cacheable:
            size = len(value)
            end = addr + size
            cached = sorted(self._cache.overlap(addr, end), key=lambda x:x.begin)
            self._metrics.writes += size

            if len(cached):
                # Write data is entirely within a single cached interval.
                if addr >= cached[0].begin and end <= cached[0].end:
                    beginOffset = addr - cached[0].begin
                    endOffset = beginOffset + size
                    cached[0].data[beginOffset:endOffset] = value

                else:
                    self._update_contiguous(cached, addr, bytearray(value))
            else:
                # No cached data in this range, so just add the entire interval.
                self._cache.addi(addr, end, bytearray(value))

        return result

    def write_memory_block32(self, addr, data):
        return self.write_memory_block8(addr, conversion.u32le_list_to_byte_list(data))

    def invalidate(self):
        self._reset_cache()




# pyOCD debugger
# Copyright (c) 2016-2020 Arm Limited
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .provider import (TargetThread, ThreadProvider)
from .common import (read_c_string, HandlerModeThread, EXC_RETURN_EXT_FRAME_MASK)
from ..core import exceptions
from ..core.target import Target
from ..core.plugin import Plugin
from ..debug.context import DebugContext
from ..coresight.cortex_m_core_registers import index_for_reg
import logging

FREERTOS_MAX_PRIORITIES	= 63

LIST_SIZE = 20
LIST_INDEX_OFFSET = 16
LIST_NODE_NEXT_OFFSET = 8 # 4?
LIST_NODE_OBJECT_OFFSET = 12

THREAD_STACK_POINTER_OFFSET = 0
THREAD_PRIORITY_OFFSET = 44
THREAD_NAME_OFFSET = 52

# Create a logger for this module.
LOG = logging.getLogger(__name__)

class TargetList(object):
    def __init__(self, context, ptr):
        self._context = context
        self._list = ptr

    def __iter__(self):
        prev = -1
        found = 0
        count = self._context.read32(self._list)
        if count == 0:
            return

        node = self._context.read32(self._list + LIST_INDEX_OFFSET)

        while (node != 0) and (node != prev) and (found < count):
            try:
                # Read the object from the node.
                obj = self._context.read32(node + LIST_NODE_OBJECT_OFFSET)
                yield obj
                found += 1

                # Read next list node pointer.
                prev = node
                node = self._context.read32(node + LIST_NODE_NEXT_OFFSET)
            except exceptions.TransferError:
                LOG.warning("TransferError while reading list elements (list=0x%08x, node=0x%08x), terminating list", self._list, node)
                node = 0

class FreeRTOSThreadContext(DebugContext):
    """! @brief Thread context for FreeRTOS."""
    
    # SP/PSP are handled specially, so it is not in these dicts.

    COMMON_REGISTER_OFFSETS = {
                 4: 0, # r4
                 5: 4, # r5
                 6: 8, # r6
                 7: 12, # r7
                 8: 16, # r8
                 9: 20, # r9
                 10: 24, # r10
                 11: 28, # r11
            }

    NOFPU_REGISTER_OFFSETS = {
                 0: 32, # r0
                 1: 36, # r1
                 2: 40, # r2
                 3: 44, # r3
                 12: 48, # r12
                 14: 52, # lr
                 15: 56, # pc
                 16: 60, # xpsr
            }
    NOFPU_REGISTER_OFFSETS.update(COMMON_REGISTER_OFFSETS)

    FPU_BASIC_REGISTER_OFFSETS = {
                -1: 32, # exception LR
                 0: 36, # r0
                 1: 40, # r1
                 2: 44, # r2
                 3: 48, # r3
                 12: 52, # r12
                 14: 56, # lr
                 15: 60, # pc
                 16: 64, # xpsr
            }
    FPU_BASIC_REGISTER_OFFSETS.update(COMMON_REGISTER_OFFSETS)

    FPU_EXTENDED_REGISTER_OFFSETS = {
                -1: 32, # exception LR
                 0x50: 36, # s16
                 0x51: 40, # s17
                 0x52: 44, # s18
                 0x53: 48, # s19
                 0x54: 52, # s20
                 0x55: 56, # s21
                 0x56: 60, # s22
                 0x57: 64, # s23
                 0x58: 68, # s24
                 0x59: 72, # s25
                 0x5a: 76, # s26
                 0x5b: 80, # s27
                 0x5c: 84, # s28
                 0x5d: 88, # s29
                 0x5e: 92, # s30
                 0x5f: 96, # s31
                 0: 100, # r0
                 1: 104, # r1
                 2: 108, # r2
                 3: 112, # r3
                 12: 116, # r12
                 14: 120, # lr
                 15: 124, # pc
                 16: 128, # xpsr
                 0x40: 132, # s0
                 0x41: 136, # s1
                 0x42: 140, # s2
                 0x43: 144, # s3
                 0x44: 148, # s4
                 0x45: 152, # s5
                 0x46: 156, # s6
                 0x47: 160, # s7
                 0x48: 164, # s8
                 0x49: 168, # s9
                 0x4a: 172, # s10
                 0x4b: 176, # s11
                 0x4c: 180, # s12
                 0x4d: 184, # s13
                 0x4e: 188, # s14
                 0x4f: 192, # s15
                 33: 196, # fpscr
                 # (reserved word: 200)
            }
    FPU_EXTENDED_REGISTER_OFFSETS.update(COMMON_REGISTER_OFFSETS)

    def __init__(self, parent, thread):
        super(FreeRTOSThreadContext, self).__init__(parent)
        self._thread = thread
        self._has_fpu = self.core.has_fpu

    def read_core_registers_raw(self, reg_list):
        reg_list = [index_for_reg(reg) for reg in reg_list]
        reg_vals = []

        isCurrent = self._thread.is_current
        inException = isCurrent and self._parent.read_core_register('ipsr') > 0

        # If this is the current thread and we're not in an exception, just read the live registers.
        if isCurrent and not inException:
            return self._parent.read_core_registers_raw(reg_list)

        # Because of above tests, from now on, inException implies isCurrent;
        # we are generating the thread view for the RTOS thread where the
        # exception occurred; the actual Handler Mode thread view is produced
        # by HandlerModeThread
        if inException:
            # Reasonable to assume PSP is still valid
            sp = self._parent.read_core_register('psp')
        else:
            sp = self._thread.get_stack_pointer()

        # Determine which register offset table to use and the offsets past the saved state.
        hwStacked = 0x20
        swStacked = 0x20
        table = self.NOFPU_REGISTER_OFFSETS
        if self._has_fpu:
            try:
                if inException and self.core.is_vector_catch():
                    # Vector catch has just occurred, take live LR
                    exceptionLR = self._parent.read_core_register('lr')
                else:
                    # Read stacked exception return LR.
                    offset = self.FPU_BASIC_REGISTER_OFFSETS[-1]
                    exceptionLR = self._parent.read32(sp + offset)

                # Check bit 4 of the saved exception LR to determine if FPU registers were stacked.
                if (exceptionLR & EXC_RETURN_EXT_FRAME_MASK) != 0:
                    table = self.FPU_BASIC_REGISTER_OFFSETS
                    swStacked = 0x24
                else:
                    table = self.FPU_EXTENDED_REGISTER_OFFSETS
                    hwStacked = 0x68
                    swStacked = 0x64
            except exceptions.TransferError:
                LOG.debug("Transfer error while reading thread's saved LR")

        for reg in reg_list:
            # Must handle stack pointer specially.
            if reg == 13:
                if inException:
                    reg_vals.append(sp + hwStacked)
                else:
                    reg_vals.append(sp + swStacked + hwStacked)
                continue

            # Look up offset for this register on the stack.
            spOffset = table.get(reg, None)
            if spOffset is None:
                reg_vals.append(self._parent.read_core_register_raw(reg))
                continue
            if inException:
                spOffset -= swStacked

            try:
                if spOffset >= 0:
                    reg_vals.append(self._parent.read32(sp + spOffset))
                else:
                    # Not available - try live one
                    reg_vals.append(self._parent.read_core_register_raw(reg))
            except exceptions.TransferError:
                reg_vals.append(0)

        return reg_vals

class FreeRTOSThread(TargetThread):
    """! @brief A FreeRTOS task."""

    RUNNING = 1
    READY = 2
    BLOCKED = 3
    SUSPENDED = 4
    DELETED = 5

    STATE_NAMES = {
            RUNNING : "Running",
            READY : "Ready",
            BLOCKED : "Blocked",
            SUSPENDED : "Suspended",
            DELETED : "Deleted",
        }

    def __init__(self, targetContext, provider, base):
        super(FreeRTOSThread, self).__init__()
        self._target_context = targetContext
        self._provider = provider
        self._base = base
        self._state = FreeRTOSThread.READY
        self._thread_context = FreeRTOSThreadContext(self._target_context, self)

        self._priority = self._target_context.read32(self._base + THREAD_PRIORITY_OFFSET)

        self._name = read_c_string(self._target_context, self._base + THREAD_NAME_OFFSET)
        if len(self._name) == 0:
            self._name = "Unnamed"

    def get_stack_pointer(self):
        # Get stack pointer saved in thread struct.
        try:
            return self._target_context.read32(self._base + THREAD_STACK_POINTER_OFFSET)
        except exceptions.TransferError:
            LOG.debug("Transfer error while reading thread's stack pointer @ 0x%08x", self._base + THREAD_STACK_POINTER_OFFSET)
            return 0

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value

    @property
    def priority(self):
        return self._priority

    @property
    def unique_id(self):
        return self._base

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return "%s; Priority %d" % (self.STATE_NAMES[self.state], self.priority)

    @property
    def is_current(self):
        return self._provider.get_actual_current_thread_id() == self.unique_id

    @property
    def context(self):
        return self._thread_context

    def __str__(self):
        return "<FreeRTOSThread@0x%08x id=%x name=%s>" % (id(self), self.unique_id, self.name)

    def __repr__(self):
        return str(self)

class FreeRTOSThreadProvider(ThreadProvider):
    """! @brief Thread provider for FreeRTOS."""

    ## Required FreeRTOS symbols.
    FREERTOS_SYMBOLS = [
        "uxCurrentNumberOfTasks",
        "pxCurrentTCB",
        "pxReadyTasksLists",
        "xDelayedTaskList1",
        "xDelayedTaskList2",
        "xPendingReadyList",
        "uxTopReadyPriority",
        "xSchedulerRunning",
        ]

    def __init__(self, target):
        super(FreeRTOSThreadProvider, self).__init__(target)
        self._symbols = None
        self._total_priorities = 0
        self._threads = {}

    def init(self, symbolProvider):
        # Lookup required symbols.
        self._symbols = self._lookup_symbols(self.FREERTOS_SYMBOLS, symbolProvider)
        if self._symbols is None:
            return False

        # Look up optional xSuspendedTaskList, controlled by INCLUDE_vTaskSuspend
        suspendedTaskListSym = self._lookup_symbols(["xSuspendedTaskList"], symbolProvider)
        if suspendedTaskListSym is not None:
            self._symbols['xSuspendedTaskList'] = suspendedTaskListSym['xSuspendedTaskList']

        # Look up optional xTasksWaitingTermination, controlled by INCLUDE_vTaskDelete
        tasksWaitingTerminationSym = self._lookup_symbols(["xTasksWaitingTermination"], symbolProvider)
        if tasksWaitingTerminationSym is not None:
            self._symbols['xTasksWaitingTermination'] = tasksWaitingTerminationSym['xTasksWaitingTermination']

        # Look up vPortEnableVFP() to determine if the FreeRTOS port supports the FPU.
        vPortEnableVFP = self._lookup_symbols(["vPortEnableVFP"], symbolProvider)
        self._fpu_port = vPortEnableVFP is not None

        elfOptHelp = " Try using the --elf option." if self._target.elf is None else ""

        # Check for the expected list size. These two symbols are each a single list and xDelayedTaskList2
        # immediately follows xDelayedTaskList1, so we can just subtract their addresses to get the
        # size of a single list.
        delta = self._symbols['xDelayedTaskList2'] - self._symbols['xDelayedTaskList1']
        delta = self._get_elf_symbol_size('xDelayedTaskList1', self._symbols['xDelayedTaskList1'], delta)
        if delta != LIST_SIZE:
            LOG.warning("FreeRTOS: list size is unexpected, maybe an unsupported configuration of FreeRTOS." + elfOptHelp)
            return False

        # xDelayedTaskList1 immediately follows pxReadyTasksLists, so subtracting their addresses gives
        # us the total size of the pxReadyTaskLists array. But not trustworthy. Compiler can rearrange things
        delta = self._symbols['xDelayedTaskList1'] - self._symbols['pxReadyTasksLists']
        delta = self._get_elf_symbol_size('pxReadyTasksLists', self._symbols['pxReadyTasksLists'], delta);
        if delta % LIST_SIZE:
            LOG.warning("FreeRTOS: pxReadyTasksLists size is unexpected, maybe an unsupported version of FreeRTOS." + elfOptHelp)
            return False
        self._total_priorities = delta // LIST_SIZE
        if self._total_priorities > FREERTOS_MAX_PRIORITIES:
            LOG.warning("FreeRTOS: number of priorities is too large (%d)." + elfOptHelp, self._total_priorities)
            return False
        LOG.debug("FreeRTOS: number of priorities is %d", self._total_priorities)

        self._target.session.subscribe(self.event_handler, Target.Event.POST_FLASH_PROGRAM)
        self._target.session.subscribe(self.event_handler, Target.Event.POST_RESET)

        return True

    def invalidate(self):
        self._threads = {}

    def event_handler(self, notification):
        # Invalidate threads list if flash is reprogrammed.
        LOG.debug("FreeRTOS: invalidating threads list: %s" % (repr(notification)))
        self.invalidate();

    def _build_thread_list(self):
        newThreads = {}

        # Read the number of threads.
        threadCount = self._target_context.read32(self._symbols['uxCurrentNumberOfTasks'])

        # Read the current thread.
        currentThread = self._target_context.read32(self._symbols['pxCurrentTCB'])

        # We should only be building the thread list if the scheduler is running, so a zero thread
        # count or a null current thread means something is bizarrely wrong.
        if threadCount == 0 or currentThread == 0:
            LOG.warning("FreeRTOS: no threads even though the scheduler is running")
            return

        # Read the top ready priority.
        topPriority = self._target_context.read32(self._symbols['uxTopReadyPriority'])

        # Handle an uxTopReadyPriority value larger than the number of lists. This is most likely
        # caused by the configUSE_PORT_OPTIMISED_TASK_SELECTION option being enabled, which treats
        # uxTopReadyPriority as a bitmap instead of integer. This is ok because uxTopReadyPriority
        # in optimised mode will always be >= the actual top priority.
        if topPriority >= self._total_priorities:
            topPriority = self._total_priorities - 1

        # Build up list of all the thread lists we need to scan.
        listsToRead = []
        for i in range(topPriority + 1):
            listsToRead.append((self._symbols['pxReadyTasksLists'] + i * LIST_SIZE, FreeRTOSThread.READY))

        listsToRead.append((self._symbols['xDelayedTaskList1'], FreeRTOSThread.BLOCKED))
        listsToRead.append((self._symbols['xDelayedTaskList2'], FreeRTOSThread.BLOCKED))
        listsToRead.append((self._symbols['xPendingReadyList'], FreeRTOSThread.READY))
        if 'xSuspendedTaskList' in self._symbols:
            listsToRead.append((self._symbols['xSuspendedTaskList'], FreeRTOSThread.SUSPENDED))
        if 'xTasksWaitingTermination' in self._symbols:
            listsToRead.append((self._symbols['xTasksWaitingTermination'], FreeRTOSThread.DELETED))

        for listPtr, state in listsToRead:
            for threadBase in TargetList(self._target_context, listPtr):
                try:
                    # Don't try adding more threads than the number of threads that FreeRTOS says there are.
                    if len(newThreads) >= threadCount:
                        break

                    # Reuse existing thread objects.
                    if threadBase in self._threads:
                        t = self._threads[threadBase]
                    else:
                        t = FreeRTOSThread(self._target_context, self, threadBase)

                    # Set thread state.
                    if threadBase == currentThread:
                        t.state = FreeRTOSThread.RUNNING
                    else:
                        t.state = state

                    LOG.debug("Thread 0x%08x (%s)", threadBase, t.name)
                    newThreads[t.unique_id] = t
                except exceptions.TransferError:
                    LOG.debug("TransferError while examining thread 0x%08x", threadBase)

        if len(newThreads) != threadCount:
            LOG.warning("FreeRTOS: thread count mismatch")

        # Create fake handler mode thread.
        if self._target_context.read_core_register('ipsr') > 0:
            LOG.debug("FreeRTOS: creating handler mode thread")
            t = HandlerModeThread(self._target_context, self)
            newThreads[t.unique_id] = t

        self._threads = newThreads

    def get_threads(self):
        if not self.is_enabled:
            return []
        self.update_threads()
        return list(self._threads.values())

    def get_thread(self, threadId):
        if not self.is_enabled:
            return None
        self.update_threads()
        return self._threads.get(threadId, None)

    @property
    def is_enabled(self):
        return self._symbols is not None and self.get_is_running()

    @property
    def current_thread(self):
        if not self.is_enabled:
            return None
        self.update_threads()
        id = self.get_current_thread_id()
        try:
            return self._threads[id]
        except KeyError:
            return None

    def is_valid_thread_id(self, threadId):
        if not self.is_enabled:
            return False
        self.update_threads()
        return threadId in self._threads

    def get_current_thread_id(self):
        if not self.is_enabled:
            return None
        if self._target_context.read_core_register('ipsr') > 0:
            return HandlerModeThread.UNIQUE_ID
        return self.get_actual_current_thread_id()

    def get_actual_current_thread_id(self):
        if not self.is_enabled:
            return None
        return self._target_context.read32(self._symbols['pxCurrentTCB'])

    def get_is_running(self):
        if self._symbols is None:
            return False
        return self._target_context.read32(self._symbols['xSchedulerRunning']) != 0

    def _get_elf_symbol_size(self, name, addr, calculated_size):
        if self._target.elf is not None:
            symInfo = None
            try:
                symInfo = self._target.elf.symbol_decoder.get_symbol_for_name(name)
            except RuntimeError as e:
                LOG.error("FreeRTOS elf symbol query failed for (%s) with an exception. " + str(e),
                    name, exc_info=self._target.session.log_tracebacks)

            # Simple checks to make sure gdb is looking at the same executable we are
            if symInfo is None:
                LOG.debug("FreeRTOS symbol '%s' not found in elf file", name)
            elif symInfo.address != addr:
                LOG.debug("FreeRTOS symbol '%s' address mismatch elf=0x%08x, gdb=0x%08x", name, symInfo.address, addr)
            else:
                if calculated_size != symInfo.size:
                    LOG.info("FreeRTOS symbol '%s' size from elf (%ld) != calculated size (%ld). Using elf value.",
                        name, symInfo.size, calculated_size)
                else:
                    LOG.debug("FreeRTOS symbol '%s' size (%ld) from elf file matches calculated value", name, calculated_size)
                return symInfo.size
        return calculated_size

class FreeRTOSPlugin(Plugin):
    """! @brief Plugin class for FreeRTOS."""
    
    def load(self):
        return FreeRTOSThreadProvider
    
    @property
    def name(self):
        return "freertos"
    
    @property
    def description(self):
        return "FreeRTOS"
